/*
 * arealight.cpp - Methods that implement the area light as an object
 */

#include "arealight.h"
#include "csg.h"
#include "vector3D.h"

#include <math.h>

// Constructor
AreaLight::AreaLight( Surface* s, int orient, double m1, double m2, double M1, double M2, double m ) : Primitive(s)
{
	orientation = orient ;
	min1 = m1 ; min2 = m2 ; max1 = M1 ; max2 = M2 ; main = m ;
}

// Copy constructor
AreaLight::AreaLight( const AreaLight& a ) : Primitive()
{
	int i, j ;
	surface = a.surface ;
	orientation = a.orientation ;
	min1 = a.min1 ; min2 = a.min2 ; max1 = a.max1 ; max2 = a.max2 ; main = a.main ; 
	isTransformed = a.isTransformed ;
	
	if( a.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = a.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = a.N[i][j] ;
			}
	}
}
	
// Implements ray - area light intersection quickly
bool AreaLight::intersect( Ray& r ) 
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;
	
	double t, coord1, coord2 ; // Cyrus Beck again !
	switch( orientation )
	{
		case POSITIVE_X:
		case NEGATIVE_X:
			t = ( main - ray.origin.x ) / ray.direction.x ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.y + t * ray.direction.y ;
			coord2 = ray.origin.z + t * ray.direction.z ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			ray.length = t ; 
			break ;
		case POSITIVE_Y:
		case NEGATIVE_Y:
			t = ( main - ray.origin.y ) / ray.direction.y ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.z + t * ray.direction.z ;
			coord2 = ray.origin.x + t * ray.direction.x ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			ray.length = t ; 
			break ;
		case POSITIVE_Z:
		case NEGATIVE_Z:
			t = ( main - ray.origin.z ) / ray.direction.z ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.x + t * ray.direction.x ;
			coord2 = ray.origin.y + t * ray.direction.y ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			ray.length = t ; 
			break ;
	}

	ray.object = r.object = this ;
	// Usual retransformation code
	Vector3D point_of_intersection( ray.origin.x + ray.direction.x * ray.length,
					ray.origin.y + ray.direction.y * ray.length,
					ray.origin.z + ray.direction.z * ray.length 
				);
	if( !isTransformed ) r.length = ray.length ;
	else {
		Vector3D actual_point = transform_point( point_of_intersection ) ;
		r.length = ( Vector3D( actual_point.x - r.origin.x, 
				     actual_point.y - r.origin.y,
				     actual_point.z - r.origin.z 
				     ).length()
			   ) ;
	}
	
	// No texture mapping allowed, for obvious reasons
	return true ;
}

// Implements shadow ray - area light intersection very quickly
bool AreaLight::shadow_intersect( Ray& r ) 
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;
	
	double t, coord1, coord2 ; // Cyrus Beck again !
	switch( orientation )
	{
		case POSITIVE_X:
		case NEGATIVE_X:
			t = ( main - ray.origin.x ) / ray.direction.x ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.y + t * ray.direction.y ;
			coord2 = ray.origin.z + t * ray.direction.z ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			else return true ;
		case POSITIVE_Y:
		case NEGATIVE_Y:
			t = ( main - ray.origin.y ) / ray.direction.y ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.z + t * ray.direction.z ;
			coord2 = ray.origin.x + t * ray.direction.x ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			else return true ;
		case POSITIVE_Z:
		case NEGATIVE_Z:
			t = ( main - ray.origin.z ) / ray.direction.z ;
			if( t < 0 || t > ray.length ) return false ;
			coord1 = ray.origin.x + t * ray.direction.x ;
			coord2 = ray.origin.y + t * ray.direction.y ;
			if( coord1 < min1 || coord1 > max1 || coord2 < min2 || coord2 > max2 ) return false ;
			// Intersects successfully
			else return true ;
	}
	
	return false;
}

// Shader for an area light
Color AreaLight::shade( const Ray& ray, int depth )
{
	// Assume ray has already been intersected with the sphere - otherwise this will NOT be called
	Vector3D intersection(	ray.origin.x + ( ray.length * ray.direction.x ),
			      	ray.origin.y + ( ray.length * ray.direction.y ),
				ray.origin.z + ( ray.length * ray.direction.z )
			);
	
	Vector3D old_normal, normal ;
	switch( orientation ) 
	{
		case POSITIVE_X: old_normal.x = 1 ; old_normal.y = 0 ; old_normal.z = 0 ; break ;
		case POSITIVE_Y: old_normal.x = 0 ; old_normal.y = 1 ; old_normal.z = 0 ; break ;
		case POSITIVE_Z: old_normal.x = 0 ; old_normal.y = 0 ; old_normal.z = 1 ; break ;
		case NEGATIVE_X: old_normal.x = -1 ; old_normal.y = 0 ; old_normal.z = 0 ; break ;
		case NEGATIVE_Y: old_normal.x = 0 ; old_normal.y = -1 ; old_normal.z = 0 ; break ;
		case NEGATIVE_Z: old_normal.x = 0 ; old_normal.y = 0 ; old_normal.z = -1 ; break ;
	}
	if( isTransformed ) {
		// Transform intersection using inverse transform
		normal = transform_normal( old_normal ) ;
	}
	else normal = old_normal ;
	
	Vector3D view( -ray.direction.x, -ray.direction.y, -ray.direction.z ) ;
	
	// no textures
	return surface -> shade( intersection, normal, view, ray.n, ray.inside, depth ) ;
}

// Returns an interval of intersection of the ray r with this area
vector <CSG_Interval*> AreaLight::compute_intervals( const Ray& r )
{
	// Disabled
	vector <CSG_Interval*> result ;
	return result ;
}

void AreaLight::computeBounds( bVolume& b ) // Compute bounding volume for this area
{
	switch( orientation )
	{
		case POSITIVE_X:
		case NEGATIVE_X:
				b.ymin = min1 - DELTA ;
				b.zmin = min2 - DELTA ;
				b.ymax = max1 + DELTA ;
				b.zmax = max2 + DELTA ;
				b.xmin = main - DELTA ;
				b.xmax = main + DELTA ;
				break ;
		case POSITIVE_Y:
		case NEGATIVE_Y:
				b.zmin = min1 - DELTA ;
				b.xmin = min2 - DELTA ;
				b.zmax = max1 + DELTA ;
				b.xmax = max2 + DELTA ;
				b.ymin = main - DELTA ;
				b.ymax = main + DELTA ;
				break ;
		case POSITIVE_Z:
		case NEGATIVE_Z:
				b.xmin = min1 - DELTA ;
				b.ymin = min2 - DELTA ;
				b.xmax = max1 + DELTA ;
				b.ymax = max2 + DELTA ;
				b.zmin = main - DELTA ;
				b.zmax = main + DELTA ;
				break ;
	}
	
	if( isTransformed ) {
		// Transform each of the vertices of the current bounding box using the same transformation
		// Take min x, min y, min z, max x, max y, max z
		transform_box( b ) ;
	}
}
