// box.cpp - Implements intersection and shading methods for a box

#include "box.h"
#include "csg.h"
#include "vector3D.h"

#include <math.h>

// Constructor
Box::Box( Surface* s, const Vector3D& low, const Vector3D& high ) : Primitive(s)
{
	l = low ;
	h = high ;
}

// Copy constructor
Box::Box( const Box& b ) : Primitive()
{
	int i, j ;
	surface = b.surface ;
	l = b.l ;
	h = b.h ;
	isTransformed = b.isTransformed ;
	if( b.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = b.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = b.N[i][j] ;
			}
	}
}

// Checks if point "p" is inside the box pointed to by b
bool IsPointInBox( const Vector3D& p, Box* b ) 
{
	return( b -> l.x < p.x && p.x < b -> h.x && b -> l.y < p.y && p.y < b -> h.y && b -> l.z < p.z && p.z < b -> h.z ) ;
}

bool Box::shadow_intersect( Ray& r ) 
{
	return intersect( r ) ;
}

// The crucial intersect() method - implements ray object intersection method for a box
// Determines if this box is the first along the line of sight from the eye to intersect the given ray
// If not, nothing is done otherwise the ray's object is updated to this box, which could of course again 
// be overridden by another, closer object.
// Finally, this code is modified for when rays can originate inside the box (after refraction is implemented)
// We use the boolean variable inside to indicate if the ray originated inside of a closed object
// This way, it becomes easier and faster to do the intersection calculation
bool Box::intersect( Ray& r ) 
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;

	// The intersection is computed very easily since the box is very regular in shape and has its axes aligned with the coordinate axes
	// We still need to do some sort of a Cyrus beck as in the plane case, but determining the correct plane can also be
	// a bit of an issue
	// This can be done on a case by case basis
	
	// First find if origin is inside or outside
	bool inside = IsPointInBox( ray.origin, this ) ;
	
	// Compute six values of t and take the minimum - this is the simplest method
	double t, minimum = MAX_LENGTH, x, y, z ;
	
	// x Faces
	if( DELTA > ray.direction.x && ray.direction.x > -DELTA ) ; // Parallel
	else {
		t = ( h.x - ray.origin.x ) / ray.direction.x ;
		if( t > 0 ) {
			y = ray.origin.y + t * ray.direction.y ;
			z = ray.origin.z + t * ray.direction.z ;

			if( y < l.y || y > h.y || z < l.z || z > h.z ) ;
			else
				minimum = t ;
		}
		
		t = ( l.x - ray.origin.x ) / ray.direction.x ;
		if( t > 0 && t < minimum ) {
			y = ray.origin.y + t * ray.direction.y ;
			z = ray.origin.z + t * ray.direction.z ;

			if( y < l.y || y > h.y || z < l.z || z > h.z ) ;
			else
				minimum = t ;
		}
	}
	
	// y Faces
	if( DELTA > ray.direction.y && ray.direction.y > -DELTA ) ; // Parallel
	else {
		t = ( h.y - ray.origin.y ) / ray.direction.y ;
		if( t > 0 && t < minimum ) {
			z = ray.origin.z + t * ray.direction.z ;
			x = ray.origin.x + t * ray.direction.x ;

			if( z < l.z || z > h.z || x < l.x || x > h.x ) ;
			else 
				minimum = t ;
		}
		
		t = ( l.y - ray.origin.y ) / ray.direction.y ;
		if( t > 0 && t < minimum ) {
			z = ray.origin.z + t * ray.direction.z ;
			x = ray.origin.x + t * ray.direction.x ;

			if( z < l.z || z > h.z || x < l.x || x > h.x ) ;
			else 
				minimum = t ;
		}
	}
	
	// z Faces
	if( DELTA > ray.direction.z && ray.direction.z > -DELTA ) ; // Parallel
	else {
		t = ( h.z - ray.origin.z ) / ray.direction.z ;
		if( t > 0 && t < minimum ) {
			x = ray.origin.x + t * ray.direction.x ;
			y = ray.origin.y + t * ray.direction.y ;

			if( x < l.x || x > h.x || y < l.y || y > h.y ) ;
			else 
				minimum = t ;
		}
		
		t = ( l.z - ray.origin.z ) / ray.direction.z ;
		if( t > 0 && t < minimum ) {
			x = ray.origin.x + t * ray.direction.x ;
			y = ray.origin.y + t * ray.direction.y ;

			if( x < l.x || x > h.x || y < l.y || y > h.y ) ;
			else 
				minimum = t ;
		}
	}

	if( minimum == MAX_LENGTH || minimum > ray.length )
		return false ;
	else {
		ray.length = minimum ;
		ray.object = this ;
		r.object = this ;
		ray.inside = r.inside = inside ;
		if( isTransformed ) {
			Vector3D original_point = Vector3D( ray.origin.x + minimum * ray.direction.x,
							    ray.origin.y + minimum * ray.direction.y,
							    ray.origin.z + minimum * ray.direction.z ) ;
			Vector3D actual_point = transform_point( original_point
							) ;
		r.length = ( Vector3D( actual_point.x - r.origin.x, actual_point.y - r.origin.y, actual_point.z - r.origin.z ) ).length() ;
		}

		else r.length = ray.length ;
		Vector3D intersection_pt( 
				r.origin.x + r.length * r.direction.x, 
				r.origin.y + r.length * r.direction.y, 
				r.origin.z + r.length * r.direction.z 
				);
		 //printf("\nDebug: Ray intersects box with range (%lf,%lf,%lf) to (%lf,%lf,%lf) at (%lf,%lf,%lf)\n", l.x, l.y, l.z, h.x, h.y, h.z, intersection_pt.x, intersection_pt.y, intersection_pt.z ) ;
		return true ;
	}
	return false ;
}

// Object shader for a box
Color Box::shade( const Ray& ray, int depth )
{
	double u, v ;
	if( ray.object != this && ray.intersector != this ) return Color( 0, 0, 0 ) ;
	// Assume ray has already been intersected with the box - otherwise this will NOT be called
	Vector3D intersection(	ray.origin.x + ( ray.length * ray.direction.x ),
			      	ray.origin.y + ( ray.length * ray.direction.y ),
				ray.origin.z + ( ray.length * ray.direction.z )
			);
	Vector3D actual_point, normal ;
	
	if( isTransformed ) {
		// Transform intersection using inverse transform
		double result[3] ;
		for( int i = 0 ; i < 3 ; i++ )
		result[i] = inverse[i][0] * intersection.x +
			    inverse[i][1] * intersection.y +
		      	    inverse[i][2] * intersection.z +
			    inverse[i][3] ;
		actual_point = Vector3D( result[0], result[1], result[2] ) ;
	}
	else actual_point = intersection ;

	if( isZero( actual_point.x - l.x ) ) 
	{
		normal = Vector3D( -1, 0, 0 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.y - l.y ) / ( h.y - l.y ) ;
			v = ( actual_point.z - l.z ) / ( h.z - l.z ) ;
		}
	}
	
	else if( isZero( actual_point.x - h.x ) ) 
	{
		normal = Vector3D( 1, 0, 0 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.y - l.y ) / ( h.y - l.y ) ;
			v = ( actual_point.z - l.z ) / ( h.z - l.z ) ;
		}
	}
	
	else if( isZero( actual_point.y - l.y ) ) 
	{
		normal = Vector3D( 0, -1, 0 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.z - l.z ) / ( h.z - l.z ) ;
			v = ( actual_point.x - l.x ) / ( h.x - l.x ) ;
		}
	}
	
	else if( isZero( actual_point.y - h.y ) ) 
	{
		normal = Vector3D( 0, 1, 0 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.z - l.z ) / ( h.z - l.z ) ;
			v = ( actual_point.x - l.x ) / ( h.x - l.x ) ;
		}
	}
	
	else if( isZero( actual_point.z - l.z ) ) 
	{
		normal = Vector3D( 0, 0, -1 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.x - l.x ) / ( h.x - l.x ) ;
			v = ( actual_point.y - l.y ) / ( h.y - l.y ) ;
		}
	}
	
	else if( isZero( actual_point.z - h.z ) )
	{
		normal = Vector3D( 0, 0, 1 ) ; 
		if( surface -> texture ) {
			u = ( actual_point.x - l.x ) / ( h.x - l.x ) ;
			v = ( actual_point.y - l.y ) / ( h.y - l.y ) ;
		}
	}
	
	Vector3D view( -ray.direction.x, -ray.direction.y, -ray.direction.z ) ;
	
	Vector3D surface_normal ;
	if( isTransformed ) surface_normal = transform_normal( normal ) ;
	else surface_normal = normal ;
	
	if( surface -> texture ) {
		if( u < DELTA ) u = DELTA ;
		if( v < DELTA ) v = DELTA ;
		if( u > 1 - DELTA ) u = 1 - DELTA ;
		if( v > 1 - DELTA ) v = 1 - DELTA ;
		// Call texture shader with appropriate parameters
		return surface -> texture_shade( u, v, intersection, surface_normal, view, ray.n, ray.inside, depth ) ;
	}
	
	// Call surface shader with appropriate parameters
	else return surface -> shade( intersection, surface_normal, view, ray.n, ray.inside, depth ) ;
}

// Returns an interval of intersection of the ray r with this box
vector <CSG_Interval*> Box::compute_intervals( const Ray& r ) 
{
	// Similar to the intersection routine to begin with
	Ray ray ;
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}
	
	else ray = r ;
	
	// First find if origin is inside or outside
	bool inside = IsPointInBox( ray.origin, this ) ;
	
	// Compute six values of t and take the minimum - this is the simplest method
	double t[6], dotproduct ;
	
	// x Faces
	if( fabs( ray.direction.x ) < DELTA ) t[0] = t[1] = MAX_LENGTH ; // Parallel
	else {
		t[0] = ( h.x - ray.origin.x ) / ray.direction.x ;
		t[1] = ( l.x - ray.origin.x ) / ray.direction.x ;
		Vector3D intersectionH( h.x, ray.origin.y + t[0] * ray.direction.y, ray.origin.z + t[0] * ray.direction.z );
		Vector3D intersectionL( l.x, ray.origin.y + t[1] * ray.direction.y, ray.origin.z + t[1] * ray.direction.z );

		if( l.y > intersectionH.y || intersectionH.y > h.y || intersectionH.z < l.z || intersectionH.z > h.z )
			t[0] = -1 ;
		if( l.y > intersectionL.y || intersectionL.y > h.y || intersectionL.z < l.z || intersectionL.z > h.z )
			t[1] = -1 ;
	}
	
	// y Faces
	if( fabs( ray.direction.y ) < DELTA ) t[2] = t[3] = MAX_LENGTH ; // Parallel
	else {
		t[2] = ( h.y - ray.origin.y ) / ray.direction.y ;
		t[3] = ( l.y - ray.origin.y ) / ray.direction.y ;
		Vector3D intersectionH( ray.origin.x + t[2] * ray.direction.x, h.y, ray.origin.z + t[2] * ray.direction.z );
		Vector3D intersectionL( ray.origin.x + t[3] * ray.direction.x, l.y, ray.origin.z + t[3] * ray.direction.z );

		if( l.x > intersectionH.x || intersectionH.x > h.x || intersectionH.z < l.z || intersectionH.z > h.z )
			t[2] = -1 ;
		if( l.x > intersectionL.x || intersectionL.x > h.x || intersectionL.z < l.z || intersectionL.z > h.z )
			t[3] = -1 ;
	}
	
	// z Faces
	if( fabs( ray.direction.z ) < DELTA ) t[4] = t[5] = MAX_LENGTH ; // Parallel
	else {
		t[4] = ( h.z - ray.origin.z ) / ray.direction.z ;
		t[5] = ( l.z - ray.origin.z ) / ray.direction.z ;
		Vector3D intersectionH( ray.origin.x + t[4] * ray.direction.x, ray.origin.y + t[4] * ray.direction.y, h.z );
		Vector3D intersectionL( ray.origin.x + t[5] * ray.direction.x, ray.origin.y + t[5] * ray.direction.y, l.z );

		if( l.x > intersectionH.x || intersectionH.x > h.x || intersectionH.y < l.y || intersectionH.y > h.y )
			t[4] = -1 ;
		if( l.x > intersectionL.x || intersectionL.x > h.x || intersectionL.y < l.y || intersectionL.y > h.y )
			t[5] = -1 ;
	}

	double minimum = MAX_LENGTH ; int pos = -1 ;
	for( int i = 0 ; i < 6 ; i++ ) if( t[i] < minimum && t[i] > 0 ) { minimum = t[i] ; pos = i ; }
	
	double tnear, tfar ;
	if( pos == -1 ) {
		vector < CSG_Interval* > result ;
		return result ;
	}

	if( inside ) {
		tnear = 0 ;
		tfar = minimum ;
	}

	else {
		double second_minimum = MAX_LENGTH ;
		for( int i = 0 ; i < 6 ; i++ ) if( t[i] < second_minimum && t[i] > 0 && i != pos ) second_minimum = t[i] ;
		tnear = minimum ;
		tfar = second_minimum ;
	}
	
	if( isTransformed ) {
		// Transform back !
		if( !isZero( tnear ) ) {
			Vector3D nearpoint(ray.origin.x + tnear * ray.direction.x,
					ray.origin.y + tnear * ray.direction.y,	
					ray.origin.z + tnear * ray.direction.z
					);
			Vector3D transformed_point = transform_point( nearpoint ) ;
			tnear = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}

		if( !isZero( tfar ) ) {
			Vector3D farpoint(  ray.origin.x + tfar * ray.direction.x,
					ray.origin.y + tfar * ray.direction.y,	
					ray.origin.z + tfar * ray.direction.z
					);
			// Repeat calculation for far point of intersection
			Vector3D transformed_point = transform_point( farpoint ) ; 
			tfar = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}
	}

	vector <CSG_Interval*> result ;
	CSG_Interval* new_interval = new CSG_Interval() ;
	new_interval -> tnear = tnear ; new_interval -> tfar = tfar ; 
	new_interval -> pnear = new_interval -> pfar = this ;
	result.push_back( new_interval ) ;
	return result ;
}

void Box::computeBounds( bVolume& b ) 
{
	b.xmin = l.x ; b.ymin = l.y ; b.zmin = l.z ;	
	b.xmax = h.x ; b.ymax = h.y ; b.zmax = h.z ;	
}
