// Implements axis-aligned bounding volume hierarchies to speed up ray tracing
// References: 
// Ray Tracing Complex Scenes, Kay-Kajiya, ACM SIGGRAPH, 1986
// Computer Graphics - Foley, Van Dam, Feiner and Hughes

#include "bvolume.h"
#include "ray.h"
#include "surface.h" // For definition of DELTA

#include <assert.h> // For assert()
#include <math.h> // For fabs()
#include <queue>
#include <map>

using std::queue ;
using std::map ;

extern map<Object*,int> objectMap ;

// We need two sets of methods
// One builds the hierarchy of bounding volumes given a set of objects
// The second traverses the hierarchy efficiently to compute ray-object intersections using
// the algorithm presented in Kay-Kajiya's paper

// First class / function / data declarations

// Method to insert a new bounding volume into a hierarchy
// The appropriate position for insertion of the bounding volume 
// into the hierarchy is determined so that it minimizes the Goldsmith-Salmon
// cost criterion (Reference: Foley chap 15, Efficiency Considerations for Ray Tracing)
// This is based on the (relative) probability that a ray strikes a bounding volume
// which in turn is proportional to the surface area of the bounding box
// The complexity of the insert() method implemented below is O(log n)
// and the overall cost of building the hierarchy is O( n log n )
// This preprocessing cost is hugely compensated by the savings for each ray
// that cumulate, especially for complex scenes
// Hence the relevance of this optimization
void bVolumeHierarchy::insert( bVolume B ) 
{
	// Insert this bounding volume into the hierarchy
	if( root == NULL ) {
		root = new bVolumeNode( B ) ;
		return ;
	}

	// Traverse hierarchy and keep track of cost of insertion at each point
	double mincost = MAX_LENGTH ;

	bVolumeNode* currentNode = root ; bVolumeNode* insertNode ;
	double ancestor_cost = 0 ;
	int k = currentNode -> nc ;

	queue<int> rememberQ ; // remembers each child number we visited
	
	while( k ) {
		// Calculate the incremental cost of adding the new bounding volume as a child of current Node
		// This should be the sum of the cost of increase of the ancestors' areas and the increase for this node
		double Anew = combine_area( currentNode -> box, B ) ;
		double area_incr_cost = k * ( Anew - ( currentNode -> area ) ) ;
		double newcost = ancestor_cost + area_incr_cost + Anew ;
		if( newcost < mincost ) {
			mincost = newcost ;
			insertNode = currentNode ;
		}

		// Now proceed to appropriate subtree - use a greedy strategy and choose the one with least area
		bVolumeNode* minNode = (currentNode -> childList)[0] ;
		int pos = 0 ;
		for( int i = 1 ; i < k ; i++ )
		{
			if( ( (currentNode -> childList)[i] -> area ) < ( minNode -> area ) ) {
				minNode = (currentNode -> childList)[i] ;
				pos = i ;
			}
		}

		// Increment ancestor cost correctly
		ancestor_cost += area_incr_cost ;
		
		// Proceed to child minNode
		currentNode = minNode ; rememberQ.push( pos ) ;
		k = currentNode -> nc ;
	}

	double leafcost = ancestor_cost + 2 * combine_area( currentNode -> box, B ) ;
	if( leafcost < mincost ) {
		// Make leaf and this children
		bVolumeNode* newLeaf = new bVolumeNode( currentNode -> box ) ;
		bVolumeNode* newNode = new bVolumeNode( B ) ;
		(currentNode -> childList).push_back( newLeaf ) ;
		(currentNode -> childList).push_back( newNode ) ;
		currentNode -> nc = 2 ;
		( currentNode -> box ).obj = NULL ;
		insertNode = currentNode ;
		// Update area, etc... later
	}

	else {
		// insertNode stores the best point in the hierarchy to insert the new bounding box
		// So insert at this point as a child
		// Update cost of grandparent, etc...
		bVolumeNode* newNode = new bVolumeNode( B ) ;
		(insertNode -> childList).push_back( newNode ) ;
		(insertNode -> nc)++ ;
	}
	
	// Update areas of ancestors using rememberQ upto insertNode
	ancestor_cost = 0 ;
	currentNode = root ;
	do {
		double Anew = combine_area( currentNode -> box, B ) ;
		currentNode -> area = Anew ;
		combine( currentNode -> box, currentNode -> box, B ) ;
		if( currentNode == insertNode ) break ;
		currentNode = currentNode -> childList[ rememberQ.front() ] ;
		rememberQ.pop() ;
	}while( 1 ) ; 
}

// Builds a bounding volume hierarchy from a vector of objects
void bVolumeHierarchy::buildHierarchyFromObjects( vector<Object*> objects )
{
	// Keep computing object bVolumes and adding them using insert()
	vector<Object*>::iterator it ;
	for( it = objects.begin() ; it != objects.end() ; it++ ) 
	{
		bVolume b ;
		(*it) -> computeBounds(b) ;
		b.obj = *it ;
		insert(b) ;
	}
}
		
double combine_area( bVolume b1, bVolume b2 ) 
{
	bVolume b ;
	combine( b, b1, b2 ) ;
	double x = b.xmax - b.xmin ;	
	double y = b.ymax - b.ymin ;	
	double z = b.zmax - b.zmin ;
	return ( x * y + y * z + z * x ) ;
}

void intersect_vol( bVolume& b, bVolume b1, bVolume b2 )
{
	b.xmin = max( b1.xmin, b2.xmin ) ;
	b.ymin = max( b1.ymin, b2.ymin ) ;
	b.zmin = max( b1.zmin, b2.zmin ) ;
	b.xmax = min( b1.xmax, b2.xmax ) ;
	b.ymax = min( b1.ymax, b2.ymax ) ;
	b.zmax = min( b1.zmax, b2.zmax ) ;
}

void combine( bVolume& b, bVolume b1, bVolume b2 )
{
	b.xmin = min( b1.xmin, b2.xmin ) ;
	b.ymin = min( b1.ymin, b2.ymin ) ;
	b.zmin = min( b1.zmin, b2.zmin ) ;
	b.xmax = max( b1.xmax, b2.xmax ) ;
	b.ymax = max( b1.ymax, b2.ymax ) ;
	b.zmax = max( b1.zmax, b2.zmax ) ;
}


// Optimized method for shadow - ray - bounding box intersection
bool bVolumeNode::shadow_intersect( Ray& ray )
{
	if( !nc ) {
		//assert( box.obj ) ;
		// Leaf node (primitive) - therefore simply intersect with this primitive
		return ( box.obj ) -> shadow_intersect( ray ) ;
	}

	// Interior node - more work involved akin to ray-box intersection as in box.cpp
	// However, the advantage is that we do not need to compute ray.length
	// Only need to find if there exists an intersection
	
	// Ray originates inside bounding box
	
	if( IsPointInBoundingVolume( ray.origin ) ) return true ;
	
	// Code as in box.cpp
	// Use just three values and cull back faces to speed up
	double t ; 
	double x, y, z ;
	double xo = ray.origin.x, yo = ray.origin.y, zo = ray.origin.z, xd = ray.direction.x, yd = ray.direction.y, zd = ray.direction.z ;
	
	// x Faces
	if( DELTA > xd && xd > -DELTA ) ; // Parallel
	else {
		x = ( xd < 0 ) ? box.xmax : box.xmin ;
		t = ( x - xo ) / xd ;
		if( t > 0 && t < ray.length ) {
			y = yo + t * yd ;
			z = zo + t * zd ;

			if( y < box.ymin || y > box.ymax || z < box.zmin || z > box.zmax ) ;
			else return true ;	
		}
	}
	
	// y Faces
	if( DELTA > yd && yd > -DELTA ) ; // Parallel
	else {
		y = ( yd < 0 ) ? box.ymax : box.ymin ;
		t = ( y - yo ) / yd ;
		if( t > 0 && t < ray.length ) {
			z = zo + t * zd ;
			x = xo + t * xd ;

			if( z < box.zmin || z > box.zmax || x < box.xmin || x > box.xmax ) ;
			else return true ;
		}
	}
	
	// z Faces
	if( DELTA > zd && zd > -DELTA ) return false ; // Parallel
	else {
		z = ( zd < 0 ) ? box.zmax : box.zmin ;
		t = ( z - zo ) / zd ;
		if( t > 0 && t < ray.length ) {
			x = xo + t * xd ;
			y = yo + t * yd ;

			if( x < box.xmin || x > box.xmax || y < box.ymin || y > box.ymax ) return false ;
			else return true ;
		}
		else return false ;
	}
}

inline bool bVolumeNode::IsPointInBoundingVolume( const Vector3D& p )
{
	return ( box.xmin < p.x && p.x < box.xmax && box.ymin < p.y && p.y < box.ymax && box.zmin < p.z && p.z < box.zmax ) ;
}

// Optimized method for ray - bounding box intersection
// gprof says this is the most frequently called routine and takes the most time
// Badly needs to be optimized
bool bVolumeNode::intersect( Ray& ray ) 
{
	if( !nc ) {
		// assert( box.obj ) ;
		// Leaf node (primitive) - therefore simply intersect with this primitive
		bool result = ( box.obj ) -> intersect( ray ) ;
		if( result ) distance_from_ray = ray.length ;
		else distance_from_ray = MAX_LENGTH ;
		return result ;
	}
	
	// Subtle bug when ray originates inside a bounding box
	// This is not a problem for shadow rays (Why? I now think it is!)
	bool inside = IsPointInBoundingVolume( ray.origin ) ;
	if( inside ) {
		distance_from_ray = -1 ; // Closest
		return true ;
	}
	
	// Compute six values of t and take the minimum - this is the simplest method
	// And also the slowest - hence it is scrapped herewith
	// Use just three values and cull back faces to speed up
	double tx, ty, tz, dotproduct, x, y, z, minimum ;
	bool flag = false ;	
	// x Faces
	if( DELTA > ray.direction.x && ray.direction.x > -DELTA ) tx = MAX_LENGTH ; // Parallel
	else {
		x = ( ray.direction.x < 0 ) ? box.xmax : box.xmin ;
		tx = ( x - ray.origin.x ) / ray.direction.x ;
		y = ray.origin.y + tx * ray.direction.y ;
		z = ray.origin.z + tx * ray.direction.z ;

		if( tx < 0 || y < box.ymin || y > box.ymax || z < box.zmin || z > box.zmax )
			tx = MAX_LENGTH ;
		else flag = true ;
	}
	
	// y Faces
	if( DELTA > ray.direction.y && ray.direction.y > -DELTA ) ty = MAX_LENGTH ; // Parallel
	else {
		y = ( ray.direction.y < 0 ) ? box.ymax : box.ymin ;
		ty = ( y - ray.origin.y ) / ray.direction.y ;
		z = ray.origin.z + ty * ray.direction.z ;
		x = ray.origin.x + ty * ray.direction.x ;

		if( ty < 0 || z < box.zmin || z > box.zmax || x < box.xmin || x > box.xmax )
			ty = MAX_LENGTH ;	
		else flag = true ;
	}
	
	// z Faces
	if( DELTA > ray.direction.z && ray.direction.z > -DELTA ) tz = MAX_LENGTH ; // Parallel
	else {
		z = ( ray.direction.z < 0 ) ? box.zmax : box.zmin ;
		tz = ( z - ray.origin.z ) / ray.direction.z ;
		x = ray.origin.x + tz * ray.direction.x ;
		y = ray.origin.y + tz * ray.direction.y ;

		if( tz < 0 || x < box.xmin || x > box.xmax || y < box.ymin || y > box.ymax )
			tz = MAX_LENGTH ;	
		else flag = true ;
	}
	if( !flag ) {
		distance_from_ray = MAX_LENGTH ;
		return false ;
	}
	
	minimum = min3( tx, ty, tz ) ;
	if( minimum > ray.length ) {
		distance_from_ray = MAX_LENGTH ;
		return false ;
	}

	else {
		distance_from_ray = minimum ;
		return true ;
	}
	
	return false ;
}

void bVolumeHierarchy::print() 
{
	// Use recursive method
	if( !root ) printf("Empty hierarchy.\n") ;

	else root -> print_node() ;
}

/* Recursive method for printing a node in the bounding volume hierarchy
 */
void bVolumeNode::print_node()
{
	// Use preorder traversal
	if( nc == 0 ) {
		printf("Leaf node: box = (%lf,%lf,%lf,%lf,%lf,%lf) object = %d area = %lf\n", box.xmin, box.ymin, box.zmin, box.xmax, box.ymax, box.zmax, objectMap[ box.obj ], area ) ;
	}

	else {
		printf("Internal node: %p nc = %d box = (%lf,%lf,%lf,%lf,%lf,%lf) area = %lf\n", this, nc, box.xmin, box.ymin, box.zmin, box.xmax, box.ymax, box.zmax, area ) ;
		for( int i = 0 ; i < nc ; i++ )
			(childList[i]) -> print_node() ;
	}
}

double min( double x, double y ) 
{
	return ( ( x < y ) ? x : y ) ;
}

double max( double x, double y ) 
{
	return ( ( x > y ) ? x : y ) ;
}

inline double min3( double x, double y, double z ) 
{
	double min = x ;
	if( y < min ) min = y ;
	if( z < min ) min = z ;
	return min ;
}

double max3( double x, double y, double z ) 
{
	double max = x ;
	if( y > max ) max = y ;
	if( z > max ) max = z ;
	return max ;
}
