// Definitions for bounding volumes 

#ifndef __BVOLUME_H
#define __BVOLUME_H

#include "object.h"

#include <vector>
#include <algorithm>
#include <queue>
#include <stack>

using std::vector ;
using std::stack ;

/* Some simple compare functions */

double min( double x, double y ) ;
double max( double x, double y ) ;
double min3( double x, double y, double z ) ;
double max3( double x, double y, double z ) ;

struct bVolume {
	double xmin, ymin, zmin ; 
	double xmax, ymax, zmax ;
	Object* obj ;
};

class bVolumeNode {

	public: 
		// Data first
		bVolume box ; // Stores the extent of the axis-aligned bounding box
		double area ; // Stores the surface area of this bounding box
		double distance_from_ray ; // Distance from current ray being traced

		int nc ; // Number of children for this bounding box node in the hierarchy
		vector <bVolumeNode*> childList ; // The actual children in the hierarchy

		bVolumeNode() 
		{ 
			box.xmin = box. xmax = box.ymin = box.ymax = box.zmin = box.zmax = 0 ; area = 0 ; nc = 0 ; 
			distance_from_ray = MAX_LENGTH ;
		}

		bVolumeNode( bVolume b ) 
		{
			box.xmin = b.xmin ;
			box.ymin = b.ymin ;
			box.zmin = b.zmin ;
			box.xmax = b.xmax ;
			box.ymax = b.ymax ;
			box.zmax = b.zmax ;
			box.obj  = b.obj ;
			area =  ( b.xmax - b.xmin ) * ( b.ymax - b.ymin ) + 
				( b.ymax - b.ymin ) * ( b.zmax - b.zmin ) +
				( b.zmax - b.zmin ) * ( b.xmax - b.xmin ) ;
			nc = 0 ;
			distance_from_ray = MAX_LENGTH ;
		}
		
		bool intersect( Ray& r ) ; // Method to intersect this node with a ray
		bool shadow_intersect( Ray& r ) ; // Method to intersect a bounding box with a shadow ray
		void print_node() ; // Prints this node recursively
		bool IsPointInBoundingVolume( const Vector3D & p ) ; // Checks if a point is inside a bounding volume
};

class bVolumeHierarchy {

	public:
		bVolumeNode* root ;

		bVolumeHierarchy() { root = NULL ; }
		void insert( bVolume b ) ; // Insert a bounding volume into the hierarchy
		void buildHierarchyFromObjects( vector<Object*> objects ) ; // Build a bounding volume hierarchy from the given list of objects passed as a vector
		void print() ; // Prints the hierarchy - debug purposes
};

// Comparison class for bVolumeNode
class bVolumeNodeCompare {
	public:
	bool operator () ( bVolumeNode* b1, bVolumeNode* b2 ) {
		// Compares two bounding volumes to determine which is farther from the given ray
		// Use the top of the distance stack - safe even in recursive ray tracing, not just in ray casting
		return ( (b1 -> distance_from_ray) > (b2 -> distance_from_ray) ) ;
	}
};

double combine_area( bVolume b1, bVolume b2 ) ;
void intersect_vol( bVolume& b, bVolume b1, bVolume b2 ) ;
void combine( bVolume& b, bVolume b1, bVolume b2 ) ;

#endif
