// csg.cpp - Methods for implementing constructive solid geometry in this ray tracer

#include "csg.h"
#include <math.h> // For fabs()
#include "bvolume.h"

// We start by implementing the constructors and the three methods required to build the CSG Object from pointers to primitive objects

// Constructor initializing a primitive
CSG_Object::CSG_Object( Primitive* pptr ) {
	root = new CSG_Node ;
	(root -> data).pptr = pptr ;
}

// Build a CSG object as the union of several CSG Objects passed in a vector
void CSG_Object::buildAsUnion( vector<CSG_Object*> object_list ) 
{
	// The basic algorithm is obvious
	// Create a linked list of children and make a parent node with a union operator pointing to one of the nodes in the list
	int list_size = object_list.size() ;
	if( list_size < 2 ) return ; 
	
	int i ;
	root = new CSG_Node ;
	(root -> data).operation = CSG_UNION ;
	root -> leftChild = object_list[0] -> root ;
	for( i = 0 ; i < list_size ; i++ )
		( object_list[i] -> root ) -> rightSibling = object_list[(i+1)%list_size] -> root ; 
}

// Build a CSG object as the intersection of several CSG Objects passed in a vector
void CSG_Object::buildAsIntersection( vector<CSG_Object*> object_list ) 
{
	// The basic algorithm is obvious
	// Create a linked list of children and make a parent node with a intersection operator pointing to one of the nodes in the list
	int list_size = object_list.size() ;
	if( list_size < 2 ) return ; 
	
	int i ;
	root = new CSG_Node ;
	(root -> data).operation = CSG_INTERSECTION ;
	root -> leftChild = object_list[0] -> root ;
	for( i = 0 ; i < list_size ; i++ )
		( object_list[i] -> root ) -> rightSibling = object_list[(i+1)%list_size] -> root ; 
}

// Build a CSG object as the intersection of several CSG Objects passed in a vector
void CSG_Object::buildAsDifference( vector<CSG_Object*> object_list ) 
{
	// The basic algorithm is obvious
	// Create a linked list of children and make a parent node with a difference operator pointing to one of the nodes in the list
	int list_size = object_list.size() ;
	if( list_size < 2 ) return ; 
	
	int i ;
	root = new CSG_Node ;
	(root -> data).operation = CSG_DIFFERENCE ;
	root -> leftChild = object_list[0] -> root ;
	for( i = 0 ; i < list_size ; i++ )
		( object_list[i] -> root ) -> rightSibling = object_list[(i+1)%list_size] -> root ; 
}
// The next step will be to implement translation, scaling and rotation for CSG Objects - this is done very simply by applying these transformations to all the primitives in the object

// Translates the CSG object corresponding to this CSG Node
void CSG_Node::translate( double x, double y, double z) 
{
	if( leftChild == NULL ) 
		(data.pptr) -> translate( x, y, z ) ;
	else
	{
		// Translate each of the children in the list
		CSG_Node* currentNode = leftChild ;
		do {
			currentNode -> translate( x, y, z ) ;
			currentNode = currentNode -> rightSibling ;
		}while( currentNode != leftChild ) ;
	}
}

// Scales the CSG object corresponding to this CSG Node
void CSG_Node::scale( double Sx, double Sy, double Sz) 
{
	if( leftChild == NULL ) 
		(data.pptr) -> scale( Sx, Sy, Sz ) ;
	else
	{
		// Translate each of the children in the list
		CSG_Node* currentNode = leftChild ;
		do {
			currentNode -> scale( Sx, Sy, Sz ) ;
			currentNode = currentNode -> rightSibling ;
		}while( currentNode != leftChild ) ;
	}
}

// Rotates the CSG object corresponding to this CSG Node
void CSG_Node::rotate( char axis, double theta ) 
{
	if( leftChild == NULL ) 
		(data.pptr) -> rotate( axis, theta ) ;
	else
	{
		// Translate each of the children in the list
		CSG_Node* currentNode = leftChild ;
		do {
			currentNode -> rotate( axis, theta ) ;
			currentNode = currentNode -> rightSibling ;
		}while( currentNode != leftChild ) ;
	}
}

// Freezes the CSG object corresponding to this CSG Node by computing inverse transform
void CSG_Node::compute_inverse( void )
{
	if( leftChild == NULL ) 
		(data.pptr) -> compute_inverse() ;
	else
	{
		// Translate each of the children in the list
		CSG_Node* currentNode = leftChild ;
		do {
			currentNode -> compute_inverse() ;
			currentNode = currentNode -> rightSibling ;
		}while( currentNode != leftChild ) ;
	}
}

// Sets new surface properties for the CSG object corresponding to this CSG Node 
void CSG_Node::setSurface( Surface* s )
{
	if( leftChild == NULL ) 
		(data.pptr) -> setSurface(s) ;
	else
	{
		// Translate each of the children in the list
		CSG_Node* currentNode = leftChild ;
		do {
			currentNode -> setSurface(s) ;
			currentNode = currentNode -> rightSibling ;
		}while( currentNode != leftChild ) ;
	}
}

Surface* CSG_Node::getSurface()
{
	if( leftChild == NULL ) 
		return (data.pptr) -> getSurface() ;
	else return NULL ; // The actual location of the leak FIXME
}

// Translates this CSG Object
void CSG_Object::translate( double x, double y, double z )
{
	// Simply translate the root
	root -> translate( x, y, z ) ;
}

// Scales this CSG Object
void CSG_Object::scale( double Sx, double Sy, double Sz )
{
	// Simply scale the root
	root -> scale( Sx, Sy, Sz ) ;
}

// Rotates this CSG Object
void CSG_Object::rotate( char axis, double theta )
{
	// Simply rotate the root
	root -> rotate( axis, theta ) ;
}

// Freezes this CSG object by computing inverse transform
void CSG_Object::compute_inverse()
{
	// Simply compute inverse for the root
	root -> compute_inverse() ;
}

// Sets new surface properties for this CSG object 
void CSG_Object::setSurface( Surface* s )
{
	// Simply set surface for the root
	root -> setSurface(s) ;
}

Surface* CSG_Object::getSurface() // Gets the surface associated with this object
{
	// Simply get root's surface
	// Bug - memory leak for children TODO
	return root -> getSurface() ;
}

// Simple function to determine the union of two overlapping intervals
CSG_Interval* Interval_Union( CSG_Interval* interval1, CSG_Interval* interval2 )
{
	CSG_Interval* result = new CSG_Interval() ;
	if( interval1 -> tnear < interval2 -> tnear ) {
		result -> tnear = interval1 -> tnear ;
		result -> tfar = interval2 -> tfar ;
		result -> pnear = interval1 -> pnear ;
		result -> pfar = interval2 -> pfar ;
	}
	
	else {
		result -> tnear = interval2 -> tnear ;
		result -> tfar = interval1 -> tfar ;
		result -> pnear = interval2 -> pnear ;
		result -> pfar = interval1 -> pfar ;
	}

	return result ;
}

// Function to determine the union of a vector of interval lists, each of which is a vector
// This is a k way merging problem and can be solved in O( n log k ) time using heaps
// Remember that k is unlikely to be much larger than log k, so a better method seems to be determine the mimimum each time
// This will affect only theoretical complexity, not performance in practice
// In fact, in most cases k = 2 !
vector <CSG_Interval*> Interval_List_Union( vector < vector <CSG_Interval*> > list )
{
	// The idea is very simple for the union - simply merge the given lists
	// This may be done in two passes: 
	// * First pass - sort all intervals in all the lists by increasing order of left endpoint
	// * Second pass - Make a pass combining overlapping intervals
	// But both can be combined into a single pass if one is careful enough
	// This is what is done here
	
	// Vector to store the result
	vector <CSG_Interval*> result ;
	// As in merge, keep an array of indices to the elements
	int m = list.size() ;
	//printf("\nDebug: List size = %d\n", m ) ;
	int* n = new int[ m ] ;
	int* position = new int[ m ] ;
	int resultPosition = 0 ;
	int i, j ;
	for( i = 0 ; i < m ; i++ ) {
		position[i] = 0 ; n[i] = list[i].size() ; 
	}
	
	int currentMinPosition = 0 ;
	while( true ) {
		currentMinPosition = -1 ;
		double min = MAX_LENGTH ;
		for( i = 0 ; i < m ; i++ ) {
			if( position[i] >= n[i] ) continue ;
			double leftend = list[i][position[i]] -> tnear ;
			if( leftend < min ) {
				min = leftend ;	
				currentMinPosition = i ;
			}
		}
		
		if( currentMinPosition == -1 ) {
			//printf("\nDebug: All lists exhausted.\n") ;
			break ; // Nothing left in any of the lists
		}

		// else printf("\nDebug: Extracted minimum interval at position %d with left endpoint %f\n", currentMinPosition, min ) ;
		// Now add the interval with leftmost endpoint to the existing result
		// Merging if necessary
		CSG_Interval* candidate = list[ currentMinPosition ][ position[ currentMinPosition ] ] ;
		( position[ currentMinPosition ] )++ ;

		if( resultPosition == 0 || result[ resultPosition - 1 ] -> tfar < candidate -> tnear ) {
			//printf("\nNew interval being added\n") ;
			result.push_back( candidate );
			resultPosition++ ;
		}

		else {
			// Overlap - therefore update ONLY if necessary 
			//printf("\nNew interval not being added\n") ;
			if( result[ resultPosition - 1 ] -> tfar < candidate -> tfar )
				result[ resultPosition - 1 ] -> tfar = candidate -> tfar ;
			// Do not increment resultPosition in this case, of course !
		}
	} // end of while
	
	return result ; // This should do the trick
}

// Function to determine the complement of a list of intervals
// Used for intersection and difference operations on lists of intervals
vector <CSG_Interval*> Interval_List_Complement( vector <CSG_Interval*> v ) 
{
	int i, n, resultPosition = 0 ;
	vector <CSG_Interval*> result ;
	
	n = v.size() ;
	if( n == 0 ) {
		CSG_Interval* new_interval = new CSG_Interval() ;
		new_interval -> tnear = 0 ; 
		new_interval -> tfar = MAX_LENGTH ;
		result.push_back( new_interval ) ;
		return result ;
	}

	if( !isZero( v[0] -> tnear )  ) {
		CSG_Interval* new_interval = new CSG_Interval() ;
		new_interval -> tnear = 0 ;
		new_interval -> tfar = v[0] -> tnear ;
		new_interval -> pfar = v[0] -> pnear ;
		result.push_back( new_interval ) ;
		resultPosition++ ;
	}
	
	for( i = 0 ; i < n - 1 ; i++ ) {
		CSG_Interval* new_interval = new CSG_Interval() ;
		new_interval -> tnear = v[i] -> tfar ; new_interval -> pnear = v[i] -> pfar ;
		new_interval -> tfar = v[i+1] -> tnear ; new_interval -> pfar = v[i+1] -> pnear ;
		result.push_back( new_interval ) ;
		resultPosition++ ;
	}
	
	if( fabs( v[i] -> tfar - MAX_LENGTH ) > DELTA ) {
		CSG_Interval* new_interval = new CSG_Interval() ;
		new_interval -> tnear = v[i] -> tfar ; new_interval -> pnear = v[i] -> pfar ;
		new_interval -> tfar = MAX_LENGTH ;
		result.push_back( new_interval ) ;
		resultPosition++ ;
	}
	
	return result ;
}

// Function to determine the intersection of a vector of interval lists, each of which is a vector
// This is done by using the previous routine for union and exploiting De Morgan's laws
// The complement of the intersection being equal to the union of complements
vector <CSG_Interval*> Interval_List_Intersection( vector < vector <CSG_Interval*> > list )
{
	// First compute all the complements
	// Then take their union
	// Now complement the result
	int m = list.size() ;
	for( int i = 0 ; i < m ; i++ ) {
		//printf("\nDebug: Interval %d = ( %f, %f )\n", i + 1, ( ( (list[i]).size() == 0 ) ? 0 : (list[i])[0] -> tnear ), ( ( (list[i]).size() == 0 ) ? 0 : (list[i])[0] -> tfar ) ) ;  
		//printf("\nDebug: Size before complementing = %d\n", list[i].size() ) ;	
		if( list[i].size() == 0 ) {
			vector <CSG_Interval*> result ;
			return result ;
		}
		list[i] = Interval_List_Complement( list[i] ) ;
		//printf("\nDebug: Size after complementing = %d\n", list[i].size() ) ;	
		//printf("\nDebug: Interval %d after complementing = ( %f, %f )\n", i + 1, ( ( (list[i]).size() == 0 ) ? 0 : (list[i])[0] -> tnear ), ( ( (list[i]).size() == 0 ) ? 0 : (list[i])[0] -> tfar ) ) ;  
	}
	vector < CSG_Interval* > result = Interval_List_Union( list ) ;
	result = Interval_List_Complement( result ) ;
	//printf("\nDebug: Result Interval  = ( %f, %f )\n", ( ( (result).size() == 0 ) ? 0 : (result)[0] -> tnear ), ( ( (result).size() == 0 ) ? 0 : (result)[0] -> tfar ) ) ;  
	return result ; // This should do the trick
}

// Function to determine the difference of a vector of interval lists, each of which is a vector
// Use the definition
// A - B - C - ... = A . ( !B ) . ( !C ) ...
vector <CSG_Interval*> Interval_List_Difference( vector < vector <CSG_Interval*> > list )
{
	int m = list.size() ;
	for( int i = 1 ; i < m ; i++ ) 
		list[i] = Interval_List_Complement( list[i] ) ;
	vector < CSG_Interval* > result = Interval_List_Intersection( list ) ;
	return result ; // This should do the trick
}

// The most crucial method - A method to intersect a ray with the CSG Node representing an object constructed using CSG
// This is the crux of the entire CSG implementation
// The algorithm is straightforward:
// Determine the interval list of intersection of this ray with each of the components constituting this object,
// at the highest level
// Perform the appropriate set operation on the interval lists and determine the final interval list
vector <CSG_Interval*> CSG_Node::compute_intervals( const Ray& r ) 
{
	vector < CSG_Interval* > result ;
	if( leftChild == NULL ) {
		// No children - so this is a primitive
		result = ( (data).pptr ) -> compute_intervals( r ) ;
	}

	else {
		// Perform operation on children
		vector < vector <CSG_Interval*> > interval_list_set ;
		CSG_Node* current = leftChild ;
		do {
			//printf("\nDebug: current pointer = %p\n", current ) ;
			interval_list_set.push_back( current -> compute_intervals( r ) ) ;
			current = current -> rightSibling ;
		}while( current != leftChild ) ;
		
		switch( data.operation ) {
			case CSG_UNION: 
				result = Interval_List_Union( interval_list_set ) ;
				break ;
			case CSG_INTERSECTION: 
				result = Interval_List_Intersection( interval_list_set ) ;
				break ;
			case CSG_DIFFERENCE: 
				result = Interval_List_Difference( interval_list_set ) ;
				break ;
		}	
	}

	return result ;
}

// This method recursively computes bounding boxes for a CSG Node
// The algorithm is obvious:
// Union( u, v ) => New bounding box = union( bounding vol(u), bounding vol(v) )
// Intersection ( u, v ) => New bounding box = intersect( bounding vol(u), bounding vol(v) )
// Difference ( u, v ) => New bounding box = bounding vol(u) - a bit of a simplification but still correct
void CSG_Node::computeBounds( bVolume& b )
{
		if( leftChild == NULL ) {
				// No children - so this is a primitive
				( (data.pptr) ) -> computeBounds( b ) ;
		}

		else {
			CSG_Node* current = leftChild ;
			current -> computeBounds( b ) ;
			if( data.operation == CSG_DIFFERENCE ) return ;
			current = current -> rightSibling ;
			while( current != leftChild )
			{
					bVolume childVol ;
					current -> computeBounds( childVol ) ;
					switch( data.operation ) {
							case CSG_UNION:
								combine( b, b, childVol ) ;
								break ;
							case CSG_INTERSECTION:
								intersect_vol( b, b, childVol ) ;
								break ;
					}
					current = current -> rightSibling ;
			}
		}
}
					
// Dummy for shadow rays
bool CSG_Object::shadow_intersect( Ray& r ) 
{
	return intersect( r ) ; 
}

// Method to intersect a ray with an object constructed using CSG
// Uses the function defined above to good effect
// That is, determine the intervals of intersection correctly
// If the leftmost endpoint is non zero the ray starts outside the object and hence use this value for ray.t
// If the leftmost endpoint is zero the ray starts inside the object
// So use the corresponding right endpoint - this is where the ray leaves the object
// If the interval list is empty or both endpoints are zero then return false
// As this means that the ray does NOT intersect the CSG object
bool CSG_Object::intersect( Ray& r ) 
{
	vector <CSG_Interval*> final_interval_list = root -> compute_intervals( r ) ;
	if( final_interval_list.size() == 0 ) {
		//printf("\nDebug: No finite interval for CSG object with the above ray\n") ;
		return false ;
	}
	
	double left = (final_interval_list[0]) -> tnear, right = (final_interval_list[0]) -> tfar ;
	bool leftzero = isZero(left), rightzero = isZero(right) ;
	//printf("\nDebug: left = %f right = %f\n", left, right ) ;
	if( leftzero && rightzero ) return false ;	
	if( !leftzero && left < r.length ) {
		r.object = this ;
		r.length = left ;
		r.intersector = (final_interval_list[0]) -> pnear ;
		return true ;
	}
		
	// Ray inside object
	if( leftzero && right < r.length ) {
		r.object = this ;
		r.length = right ;
		r.intersector = (final_interval_list[0]) -> pfar ;
		r.inside = true ;
		return true ;
	}
	
	return false ;
}

/* Computes bounding volumes for a CSG Object
 * This is done by recursively invoking the corresponding method for a CSG Node
 */
void CSG_Object::computeBounds( bVolume& b ) 
{ 
	root -> computeBounds( b ) ;
}

// This method is now very simple - call the shade() method for the intersector primitive
Color CSG_Object::shade( const Ray& r, int depth ) 
{
	Color result = ( r.intersector ) -> shade( r, depth ) ;
	return result ;
}

// A method to overcome the pitfalls of floating point numbers - checks if a floating point number is zero or not
bool isZero( double x )
{
	return( ( fabs(x) < DELTA ) ) ;
}
