// csg.h - Definition for class CSG (Objects created using Constructive Solid Geometry)

#ifndef __CSG_H
#define __CSG_H

#include "object.h"
#include "primitive.h"
#include "bvolume.h"

// For NULL
#include <stddef.h>

// Possible CSG Operators
#define CSG_UNION 0
#define CSG_INTERSECTION 1
#define CSG_DIFFERENCE 2

#define MAX_CSG_LIST_SIZE 10

// Union covering data stored in internal and leaf nodes
union __CSG_NODE_DATA{
	Primitive* pptr ; // Pointer to primitive - if this is not null leaf else internal node of tree
	int operation ; // One of CSG_UNION, CSG_INTERSECTION, CSG_DIFFERENCE
};

// CSG_Node - Class definition for a node in the CSG tree
// Since a node can have multiple children we use leftChild - rightSibling representation
// For details see CLR chapter on Binomial heaps
// Each set of children is organized in a singly linked circular list
class CSG_Node {
	
	public: 
	
	__CSG_NODE_DATA data ; // Data stored in the node - this is a union as explained above	
	CSG_Node* leftChild ; // Left child of this node in the tree (if any - otherwise null)
	CSG_Node* rightSibling ; // Right sibling of this node in the tree (if any - otherwise null)
		
	// Default Constructor
	CSG_Node() { data.pptr = NULL ; leftChild = NULL ; rightSibling = this ; }
	void translate( double, double, double ) ; // Translates the CSG object corresponding to this CSG Node
	void scale( double, double, double ) ; // Scales the CSG object corresponding to this CSG Node
	void rotate( char, double ) ; // Rotates the CSG object corresponding to this CSG Node
	void compute_inverse( void ) ; // Freezes the CSG object corresponding to this CSG Node by computing inverse transform
	void setSurface( Surface* ) ; // Sets new surface properties for the CSG object corresponding to this CSG Node
	Surface* getSurface() ; // Gets the surface associated with the CSG object corresponding to this CSG Node 
	vector <CSG_Interval*> compute_intervals( const Ray& ) ;
	void computeBounds( bVolume& ) ;
	
};

// CSG_Object - Class definition for objects constructed using CSG - inherited from abstract base class Object, just like Primitive
// Contains Primitive objects arranged in a tree structure as described above
class CSG_Object : public Object
{
	public:
		
	// We only need a pointer to the root of the tree
	CSG_Node* root ;

	// Default Constructor - set root to NULL
	CSG_Object() { root = NULL ; }
	
	// Some of the virtual functions promised in abstract class object are implemented here for CSG objects
	virtual bool intersect( Ray& r ) ; // Intersects this CSG object with a ray
	virtual bool shadow_intersect( Ray& r ) ; // Intersects this CSG object with a shadow ray
	virtual void computeBounds( bVolume& b ) ; // Compute bounding volumes for this CSG object
	virtual Color shade( const Ray& r, int depth ) ; // Shades this CSG object by passing parameters to the surface shader
	virtual void translate( double, double, double ) ; // Translates this CSG object
	virtual void scale( double, double, double ) ; // Scales this CSG object
	virtual void rotate( char, double ) ; // Rotates this CSG object
	virtual void compute_inverse( void ) ; // Freezes this CSG object by computing inverse transform
	virtual void setSurface( Surface* ) ; // Sets new surface properties for this CSG object
	virtual Surface* getSurface() ; // Gets the surface associated with this object

	// Some of the functions specific to CSG objects 
	// Some of these help in building the CSG object from a set of primitives
	
	// Constructor from primitive
	CSG_Object( Primitive* pptr ) ;

	// Union method - Constructs this CSG_Object as the union of CSG_Objects, all passed in a vector
	void buildAsUnion( vector<CSG_Object*> object_list ) ;

	// Intersection method - Constructs this CSG_Object as the intersection of CSG_Objects, all passed in a vector
	void buildAsIntersection( vector<CSG_Object*> object_list ) ;
	
	// Difference method - Constructs this CSG_Object as the difference of CSG_Objects, all passed in a vector
	void buildAsDifference( vector<CSG_Object*> object_list ) ;
};

// The next important definition is of an 'interval'
// The interval primitive enables us to do intersection and shading for CSG objects
// The interval maintains its left and right values of the parameter 't' for the ray as well as the relevant primitive of intersection
// If the left value of 't' is zero, indicates ray originating inside an object
class CSG_Interval {
	
	public:
	
	double tnear, tfar ;
	Primitive* pnear, *pfar ;

	CSG_Interval() { tnear = tfar = 0 ; pnear = pfar = NULL ; }

};

// Finally, remember that each ray-CSG_Node intersection will correspond to a vector of CSG_Intervals - we need efficient algorithms for union, intersection, difference of these lists of intervals

// These are implemented by functions in csg.cpp which are prototyped below

// Simple function to determine the union of two overlapping intervals
CSG_Interval* Interval_Union( CSG_Interval* interval1, CSG_Interval* interval2 ) ;

// Simple function to determine the complement of an interval
vector <CSG_Interval*> Interval_List_Complement( vector <CSG_Interval*> v ) ; 

// Function to determine the union of a vector of interval lists
vector <CSG_Interval*> Interval_List_Union( vector < vector <CSG_Interval*> > list ) ;
	
// Function to determine the intersection of a vector of interval lists
vector <CSG_Interval*> Interval_List_Intersection( vector < vector <CSG_Interval*> > list ) ;

// Function to determine the difference of a vector of interval lists
vector <CSG_Interval*> Interval_List_Difference( vector < vector <CSG_Interval*> > list ) ;

// Function to check if a doubleing point number can be taken to be zero
bool isZero( double ) ;

#endif
