/* init.c - Contains global variable declarations and initialization code for the ray tracer
 */

/* Include all the necessary headers */
#include <vector>
#include <map>
#include <math.h>
#include <string.h>
#include <stdio.h>

#include "vector3D.h"
#include "color.h"
#include "light.h"
#include "object.h"
#include "sphere.h"
#include "arealight.h"
#include "surface.h"
#include "supersample.h" // For antialiasing method(s)
#include "ppm.h" /* For ppm file writing routines */
#include "texture.h" // For texture mapping routines and definitions

using std::vector ;
using std::map ;

#define MAX_FILENAME_LENGTH 20

/* PPM Image file format defaults */

#define RED_CHANNEL 	0
#define GREEN_CHANNEL 	1
#define BLUE_CHANNEL	2

/* Some defaults for the ray tracer - these will normally be overridden in the sdl file */

#define DEFAULT_WIDTH  		1024
#define DEFAULT_HEIGHT 		768	
#define DEFAULT_FOV 		30
#define DEFAULT_EYE_X 	 	0
#define DEFAULT_EYE_Y 	 	0
#define DEFAULT_EYE_Z 		10
#define DEFAULT_LOOKAT_X 	0
#define DEFAULT_LOOKAT_Y 	0
#define DEFAULT_LOOKAT_Z 	0
#define DEFAULT_UP_X		0
#define DEFAULT_UP_Y		1
#define DEFAULT_UP_Z		0
#define DEFAULT_BG_RED		0
#define DEFAULT_BG_GREEN	0
#define DEFAULT_BG_BLUE		0
#define DEFAULT_SURFACE_RED	0.8F
#define DEFAULT_SURFACE_GREEN	0.2F
#define DEFAULT_SURFACE_BLUE	0.9F
#define DEFAULT_SURFACE_KA	0.2F
#define DEFAULT_SURFACE_KD	0.4F
#define DEFAULT_SURFACE_KS	0.4F
#define DEFAULT_SURFACE_NS	10.0F
#define DEFAULT_SURFACE_KR	0.0F
#define DEFAULT_SURFACE_KT	0.0F
#define DEFAULT_SURFACE_N	1.0F
#define DEFAULT_PATHS_PER_PIXEL 100

/* The global declarations - these are extern'd in the yacc file which, as usual (sigh) contains main() */

Vector3D eye, lookat, up, Du, Dv, Vp ; // 3D Viewing parameters ( camera model ) for ray tracer
double fov ; // Field of view of camera / eye
Color background ; // Background color
bool eye_candy = false ; // Enable / disable antialiasing
bool enable_path = false ; // Enable/ disable path tracing
int paths_per_pixel = DEFAULT_PATHS_PER_PIXEL ;
vector<Light*> lightList ; // List of light sources
vector<Object*> objectList ; // List of finite objects in the scene
vector<Object*> infiniteObjectList ; // List of infinite objects in the scene - needed for planes 
vector<AreaLight*> areaLightList ; // List of area lights used
map<Object*,int> objectMap ; // For debugging ONLY
Surface* currentSurface ; // The current surface properties - a modal attribute of the ray tracer
Image* image ; // The final image structure to be written in ppm format to the output file

// Texture filter setting extern'd in texture.cpp
int global_texture_filter_setting = NO_FILTER ;

// Bounding Box hierarchy used in the ray tracer
bVolumeHierarchy* boxTree ;
unsigned long global_object_count = 0 ; // For debugging purposes

FILE* inputFile ; // The input file to the ray tracer 
char* inputFileName ; // The name of he input file without the 'sdl' extension

int width = DEFAULT_WIDTH, height = DEFAULT_HEIGHT ; // Parameters of output image

/* The command line argument parsing routine */
void parse_command_line( int argc, char* argv[] ) 
{
	// First must be input sdl file name 
	// Second and third optional - width, height of output image

	if( argc < 2 ) {
		printf("\nError: No input file specified. Aborting...\n") ;
		exit(0) ;
	}

	else {
		char* fname = argv[1] ;
		inputFile = fopen( fname, "r" ) ;
		if( !inputFile ) 
		{
			printf("\nError: File %s not found. Aborting...\n", fname ) ;
			exit(0) ;
		}

		inputFileName = strtok( fname, "." ) ;

		if( argc > 2 )
			width = atoi( argv[2] ) ;
		if( argc > 3 ) 	
			height = atoi( argv[3] ) ;
	}
}

/* Initialize all parameters to their defaults */
void init_defaults() 
{
	eye = Vector3D( DEFAULT_EYE_X, DEFAULT_EYE_Y, DEFAULT_EYE_Z ) ;
	lookat = Vector3D( DEFAULT_LOOKAT_X, DEFAULT_LOOKAT_Y, DEFAULT_LOOKAT_Z ) ;
	up = Vector3D( DEFAULT_UP_X, DEFAULT_UP_Y, DEFAULT_UP_Z ) ;
	background = Color( DEFAULT_BG_RED, DEFAULT_BG_GREEN, DEFAULT_BG_BLUE ) ;
	fov = DEFAULT_FOV ;
	currentSurface = new PhongSurface( Color( DEFAULT_SURFACE_RED, DEFAULT_SURFACE_GREEN, DEFAULT_SURFACE_BLUE ),
			DEFAULT_SURFACE_KA, DEFAULT_SURFACE_KD, DEFAULT_SURFACE_KS, DEFAULT_SURFACE_NS,
			DEFAULT_SURFACE_KR, DEFAULT_SURFACE_KT, DEFAULT_SURFACE_N
			);
}

/* Performs view mapping transformation for the simple camera based 3D viewing model */
void view_mapping_transform() {

	Vector3D look = Vector3D(lookat.x - eye.x, lookat.y - eye.y, lookat.z - eye.z);
	Du = Vector3D::normalize(look.cross(up));
	Dv = Vector3D::normalize(look.cross(Du));
	double fl = (double)(width / (2*tan((0.5*fov)*M_PI/180)));
	Vp = Vector3D::normalize(look);
	Vp.x = Vp.x*fl - 0.5f*(width*Du.x + height*Dv.x);
	Vp.y = Vp.y*fl - 0.5f*(width*Du.y + height*Dv.y);
	Vp.z = Vp.z*fl - 0.5f*(width*Du.z + height*Dv.z);
}

/* The initialization routine - This is also extern'd in the yacc file for obvious reasons */
void init_tracer() 
{
	image = ImageCreate( width, height ) ;
	/* Perform 3D view mapping transformation */
	view_mapping_transform() ;

	// Build the bounding volume hierarchy
	boxTree = new bVolumeHierarchy() ;
	//printf("\nDebug: Object List = ") ;
	//	for( vector<Object*>::iterator it = objectList.begin() ; it != objectList.end() ; it++ )
	//		printf("%p, ", *it ) ;
	printf("\nBuilding bounding box hierarchy...\n") ;
	boxTree -> buildHierarchyFromObjects( objectList ) ;

	// Debug: Print the constructed hierarchy
	//printf("\nDebug: Printing constructed bounding box hierarchy: ") ;
	//boxTree -> print() ;
}

Color trace_pixel_color(double i, double j) 
{
	Color pixel_color ;
	Vector3D dir = Vector3D(
			i*Du.x + j*Dv.x + Vp.x,
			i*Du.y + j*Dv.y + Vp.y,
			i*Du.z + j*Dv.z + Vp.z
			);
	Ray ray(eye, dir);
	if( ray.trace() ) {
		/* Write appropriate pixel color to ppm file */
		pixel_color = ray.shade( 0 ) ;
	} 

	else {
		pixel_color = background ;
		//cout << "\nDebug: Background color = " << background.red << " " << background.green << " " << background.blue  << "\n" ;
	}

	return pixel_color ;
}	

/* Main ray tracing routine */
void traceAll() 
{
	int i, j ;
	printf("\nBeginning ray tracing...\n") ;
	if( !eye_candy ) {
		for( j = 0 ; j < height ; j++ ) {
			//if( j == 143 ) {
			printf("Ray tracing scanline %d\n", j ) ;
			for( i = 0 ; i < width ; i++ ) {
				//if( i == 206 ) {
				Color pixel_color = trace_pixel_color( i, j ) ;
				ImageSetPixel( image, i, j, RED_CHANNEL, (unsigned char)( ( 255 ) * pixel_color.red ) ) ;
				ImageSetPixel( image, i, j, GREEN_CHANNEL, (unsigned char) ( ( 255 ) * pixel_color.green ) ) ;
				ImageSetPixel( image, i, j, BLUE_CHANNEL, (unsigned char) ( ( 255 ) * pixel_color.blue ) ) ;
			}
			//}
			//}
		}
	}

	else {
		Color* horizontal_color = new Color[ width + 1 ] ;
		Color saved_color ;

		// Tricky bdy conditions - Beware of bugs!
		for( i = 0 ; i <= width ; i++ )
			horizontal_color[i] = trace_pixel_color( i - 0.5, -0.5 ) ;

		for( j = 0 ; j < height ; j++ ) {
			printf("Ray tracing scanline %d\n", j ) ;
			saved_color = horizontal_color[0] ;

			for( i = 0 ; i < width ; i++ ) {

				Color sample[4] ;
				sample[0] = saved_color ;
				sample[1] = horizontal_color[i+1] ;

				if( i == 0 ) 
				{
					horizontal_color[0] = trace_pixel_color( -0.5, j + 0.5 ) ;
				}

				horizontal_color[i+1] = trace_pixel_color( i + 0.5, j + 0.5 ) ;
				sample[2] = horizontal_color[i+1] ;
				sample[3] = horizontal_color[i] ;
				saved_color = sample[1] ;

				Pixel centre_pixel ; centre_pixel.i = i ; centre_pixel.j = j ; centre_pixel.size = 1 ;
				Color pixel_color = adaptive_subdivide( centre_pixel, sample, 0 ) ;

				ImageSetPixel( image, i, j, RED_CHANNEL, (unsigned char)( ( 255 ) * pixel_color.red ) ) ;
				ImageSetPixel( image, i, j, GREEN_CHANNEL, (unsigned char) ( ( 255 ) * pixel_color.green ) ) ;
				ImageSetPixel( image, i, j, BLUE_CHANNEL, (unsigned char) ( ( 255 ) * pixel_color.blue ) ) ;
			}		// end of inner for
		}	// end of outer for
		delete [] horizontal_color ;
	}	// end of else
}	// end of traceAll()

void pathTraceAll()
{
	// Perform path tracing
	int i, j, k ;
	for( i = 0 ; i < infiniteObjectList.size() ; i++ )
		printf("\nDebug: Plane %d = %p\n", i, infiniteObjectList[i] ) ;
	printf("\nDebug: Sphere = %p\n", objectList[1] ) ;
	printf("\nDebug: Area light = %p\n", areaLightList[0] ) ;
	printf("\nBeginning path tracing...\n") ;
	for( j = 0 ; j < height ; j++ ) {
		//if( j == 143 ) {
		printf("Path tracing scanline %d\n", j ) ;
		for( i = 0 ; i < width ; i++ ) {	
			//if( i == 206 ) {
			//printf("Path tracing sweepline %d\n", i ) ;
			Color pixel_color(0,0,0), final_color(0,0,0) ;
			for( k = 0 ; k < paths_per_pixel ; k++ )
			{
				double u = rand_double() ; 
				double v = rand_double() ; 
				pixel_color = trace_pixel_color( (1 - u) * i + u * ( i + 1 ), ( 1 - v ) * j + v * ( j + 1 ) ) ;
				final_color.red += pixel_color.red ;	
				final_color.green += pixel_color.green ;	
				final_color.blue += pixel_color.blue ;	
			}
			final_color.red /= paths_per_pixel ;
			final_color.green /= paths_per_pixel ;
			final_color.blue /= paths_per_pixel ;
			if( final_color.red > 1.0 ) final_color.red = 1.0 ;
			if( final_color.green > 1.0 ) final_color.green = 1.0 ;
			if( final_color.blue > 1.0 ) final_color.blue = 1.0 ;
			//printf("\nDebug: Final color = %lf,%lf,%lf\n", final_color.red, final_color.green, final_color.blue ) ;
			ImageSetPixel( image, i, j, RED_CHANNEL, (unsigned char)( ( 255 ) * final_color.red ) ) ;
			ImageSetPixel( image, i, j, GREEN_CHANNEL, (unsigned char) ( ( 255 ) * final_color.green ) ) ;
			ImageSetPixel( image, i, j, BLUE_CHANNEL, (unsigned char) ( ( 255 ) * final_color.blue ) ) ;
			}
		//}	
		//}
	}
}

/* Write image to ppm file */
void write_image() 
{
	char outputFileName[ MAX_FILENAME_LENGTH ] ; 
	strcpy( outputFileName, inputFileName ) ;
	strcat( outputFileName, ".ppm" ) ;
	printf("Writing image to %s\n", outputFileName ) ;
	ImageWrite( image, outputFileName ) ;
}
