// plane.cpp - Implements intersection and shading methods for a plane

#include "plane.h"
#include "csg.h"
#include "vector3D.h"

#include <math.h>

// Constructor
Plane::Plane( Surface* s, const Vector3D& p, const Vector3D& n ) : Primitive(s)
{
	point = p ;
	normal = Vector3D::normalize(n) ;
}

// Copy constructor
Plane::Plane( const Plane& p ) : Primitive()
{
	int i, j ;
	surface = p.surface ;
	normal = p.normal ;
	point = p.point ;
	isTransformed = p.isTransformed ;
	if( p.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = p.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = p.N[i][j] ;
			}
	}
}

bool Plane::shadow_intersect( Ray& r ) 
{
	return intersect( r ) ;
}

// The crucial intersect() method - implements ray object intersection method for a plane
// Determines if this plane is the first along the line of sight from the eye to intersect the given ray
// If not, nothing is done otherwise the ray's object is updated to this plane, which could of course again 
// be overridden by another, closer object.
// Finally, this code is modified for when rays can originate "inside" the plane (after refraction is implemented)
// We use the boolean variable inside to indicate if the ray originated 'inside' of a closed object
// This way, it becomes easier and faster to do the intersection calculation
bool Plane::intersect( Ray& r ) 
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;

	// Do the intersection using parametric coordinates (Cyrus-Beck in style)
	double dotproduct = normal.dot( ray.direction ) ;
	double numerator = normal.dot( Vector3D( point.x - ray.origin.x, point.y - ray.origin.y, point.z - ray.origin.z ) ) ;
	// Ray parallel to plane
	if( isZero( dotproduct ) ) 
	{
		return false ;
	}
	
	if( ( numerator < 0 && dotproduct > 0 ) || ( numerator > 0 && dotproduct < 0 ) ) {
		return false ;
	}
	
	double new_length = numerator / dotproduct ;	

	
	if( ray.length > new_length ) {
		ray.length = new_length ;
		if( isTransformed ) {
			Vector3D intersection_point( ray.origin.x + ray.length * ray.direction.x,
						     ray.origin.y + ray.length * ray.direction.y,
						     ray.origin.z + ray.length * ray.direction.z
					);
			Vector3D actual_point = transform_point( intersection_point ) ;
			r.length = ( Vector3D( r.origin.x - actual_point.x, r.origin.y - actual_point.y, r.origin.z - actual_point.z ) ).length() ;
		}
		else r.length = ray.length ;

		ray.inside = r.inside = ( numerator > 0 ) ;
		ray.object = this ;
		r.object = this ;
		return true ;
	}
	
	else return false ; // Some other object is closer
}

// Object shader for a plane
Color Plane::shade( const Ray& ray, int depth )
{
	if( isZero( ray.length - MAX_LENGTH ) ) {
		return Color( 0, 0, 0 ) ; // Dummy intersection for infinite plane
	}
	// Assume ray has already been intersected with the plane - otherwise this will NOT be called
	Vector3D intersection(	ray.origin.x + ( ray.length * ray.direction.x ),
			      	ray.origin.y + ( ray.length * ray.direction.y ),
				ray.origin.z + ( ray.length * ray.direction.z )
			);
	
	Vector3D view( -ray.direction.x, -ray.direction.y, -ray.direction.z ) ;
	
	Vector3D surface_normal ;
	if( isTransformed ) surface_normal = transform_normal( this -> normal ) ;
	else surface_normal = this -> normal ;
	
	// Texture mapping disabled for planes as of now TODO
	// Call surface shader with appropriate parameters
	return surface -> shade( intersection, surface_normal, view, ray.n, ray.inside, depth ) ;
}

// Returns an interval of intersection of the ray r with this plane
// The convention for intervals is important as far as planes are concerned because a plane is NOT a closed object
// This could lead to a problem with a CSG object involving a plane
// For consistency, this program adopts the convention that the space on the same side as the normal is outside 
// and the other side is inside the plane
vector <CSG_Interval*> Plane::compute_intervals( const Ray& r ) 
{
	// Similar to the intersection routine to begin with
	Ray ray ;
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}
	
	else ray = r ;
	
	// Do the intersection using parametric coordinates (Cyrus-Beck in style)
	double dotproduct = normal.dot( ray.direction ) ;
	double numerator = normal.dot( Vector3D( point.x - ray.origin.x, point.y - ray.origin.y, point.z - ray.origin.z ) ) ;
	if( isZero( dotproduct ) ) {
		//printf("\nDebug: Ray originating at (%f,%f,%f)in the direction (%f,%f,%f) is parallel to plane with point (%f,%f,%f)\n", r.origin.x, r.origin.y, r.origin.z, r.direction.x, r.direction.y, r.direction.z, point.x, point.y, point.z ) ;
		// Ray does not intersect plane - therefore no interval
		vector <CSG_Interval*> result ;
		if( numerator > 0 ) {
			// Inside
			CSG_Interval* new_interval = new CSG_Interval() ;
			new_interval -> tnear = 0 ; new_interval -> tfar = MAX_LENGTH ;
			result.push_back( new_interval ) ;
		}
		return result ;
	}

	double new_length = numerator / dotproduct ;	
	if( new_length < 0 ) {
		vector <CSG_Interval*> result ;
		CSG_Interval* new_interval = new CSG_Interval() ;
		new_interval -> tnear = 0 ; new_interval -> tfar = MAX_LENGTH ;
		result.push_back( new_interval ) ;
		return result ;
	}

	// Otherwise compute interval depending on whether "inside" or "outside"

	CSG_Interval* new_interval ; double tnear, tfar ;
	if( numerator < 0 ) {
		// Outside
		tnear = new_length ;
		tfar = MAX_LENGTH ;
	}
	
	else {
		// Inside
		tnear = 0 ;
		tfar = new_length ;
	}

	if( isTransformed ) {
		// Transform back !
		if( !isZero( tnear ) ) {
			Vector3D nearpoint(ray.origin.x + tnear * ray.direction.x,
					ray.origin.y + tnear * ray.direction.y,	
					ray.origin.z + tnear * ray.direction.z
					);
			Vector3D transformed_point = transform_point( nearpoint ) ;
			tnear = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}

		if( !isZero( tfar ) ) {
			Vector3D farpoint(  ray.origin.x + tfar * ray.direction.x,
					ray.origin.y + tfar * ray.direction.y,	
					ray.origin.z + tfar * ray.direction.z
					);
			// Repeat calculation for far point of intersection
			Vector3D transformed_point = transform_point( farpoint ) ; 
			tfar = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}
	}
	
	new_interval = new CSG_Interval() ;
	new_interval -> tnear = tnear ; new_interval -> tfar = tfar ; 
	new_interval -> pnear = new_interval -> pfar = this ;
	vector <CSG_Interval*> result ;
	result.push_back( new_interval ) ;
	//printf("\nDebug: Ray originating at (%f,%f,%f)in the direction (%f,%f,%f) intersects plane from %s with point (%f,%f,%f) to give interval (%f,%f)\n", r.origin.x, r.origin.y, r.origin.z, r.direction.x, r.direction.y, r.direction.z, ( tnear == 0 ) ? "inside" : "outside", point.x, point.y, point.z, tnear, tfar ) ;
	
	return result ;
}

void Plane::computeBounds( bVolume& b ) { }
