// primitive.cpp - Implements basic initialization and transforms for any primitive

#include "primitive.h"
#include "vector3D.h"
#include "color.h"
#include <math.h>
#include <assert.h> // For assert()
#include "csg.h"

// Prototype for matrix inversion function defined in inverse.cpp
extern void matrix_invert( double** M, double inverse[4][4] ) ;

Primitive::Primitive( const Primitive& p )
{
	int i, j ;
	surface = p.surface ;
	isTransformed = p.isTransformed ;
	if( p.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = p.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = p.N[i][j] ;
			}
	}
}

Primitive::Primitive( Surface* s )
{
	surface = s ;
	isTransformed = false ;
}

// Multiplies 2 4x4 matrices
void matrix_multiply( double** A, double** B, double** C, int size = 4 )
{
	int i,j,k ;
	for( i = 0 ; i < size ; i++ )
	for( j = 0 ; j < size ; j++ )
	{
		C[i][j] = 0 ; 
		for( k = 0 ; k < size ; k++ )
			C[i][j] += A[i][k] * B[k][j] ;
	}
}

// Assigns a 4x4 matrix to another
void matrix_assign( double** A, double** B, int size = 4 )
{
	int i, j ;
	for( i = 0 ; i < size ; i++ )
		for( j = 0 ; j < size ; j++ )
			A[i][j] = B[i][j] ;
}

void Primitive::translate( double X, double Y, double Z )
{
	int i, j, k ; 
	double** T, **result ;
	T = new double*[4] ; result = new double*[4] ;
	for( i = 0 ; i < 4 ; i++ )
	{
		T[i] = new double[4] ;
		result[i] = new double[4] ;
	}

	T[0][0] = T[1][1] = T[2][2] = T[3][3] = 1 ;
	T[0][1] = T[0][2] = T[1][0] = T[1][2] = T[2][0] = T[2][1] = T[3][0] = T[3][1] = T[3][2] = 0 ;
	T[0][3] = X ;
	T[1][3] = Y ; 
	T[2][3] = Z ;
	if( !isTransformed ) {
		
		isTransformed = true ;
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;

		matrix_assign( M, T ) ;
		N[0][0] = N[1][1] = N[2][2] = 1 ;
		N[0][1] = N[0][2] = N[1][0] = N[1][2] = N[2][0] = N[2][1] = 0 ;
	}
	
	else {
		// Multiply M = TM
		matrix_multiply( T, M, result ) ;
		matrix_assign( M, result ) ;
	}
	
	for( i = 0 ; i < 4 ; i++ ) 
	{
		delete [] T[i] ; delete [] result[i] ;
	}

	delete [] T ; delete [] result ;
	// Normal does not transform as it is a line
}	

void Primitive::scale( double Sx, double Sy, double Sz )
{
	int i, j, k ; 
	double** T, **result, **NT ;
	T = new double*[4] ; result = new double*[4] ; NT = new double*[3] ;
	for( i = 0 ; i < 4 ; i++ )
	{
		T[i] = new double[4] ;
		result[i] = new double[4] ;
		if( i != 3 ) NT[i] = new double[3] ;
	}

	T[0][0] = Sx ; T[1][1] = Sy ; T[2][2] = Sz ; T[3][3] = 1 ;
	T[0][1] = T[0][2] = T[1][0] = T[1][2] = T[2][0] = T[2][1] = T[3][0] = T[3][1] = T[3][2] = 0 ;
	T[0][3] = 0 ;
	T[1][3] = 0 ; 
	T[2][3] = 0 ;
	NT[0][0] = 1 / Sx ; NT[1][1] = 1 / Sy ; NT[2][2] = 1 / Sz ;
	NT[0][1] = NT[0][2] = NT[1][0] = NT[1][2] = NT[2][0] = NT[2][1] = 0 ;
	if( !isTransformed ) {
		isTransformed = true ;
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		matrix_assign( M, T ) ;
		matrix_assign( N, NT, 3 ) ;
	}
	
	else {
		// Multiply M = TM
		matrix_multiply( T, M, result ) ;
		matrix_assign( M, result ) ;
		matrix_multiply( NT, N, result, 3 ) ;
	        matrix_assign( N, result, 3 ) ;	
	}
	
	for( i = 0 ; i < 4 ; i++ ) 
	{
		delete [] T[i] ; delete [] result[i] ;
		if( i != 3 ) delete [] NT[i] ;
	}

	delete [] T ; delete [] result ; delete [] NT ;
}
	
// Rotate routine
void Primitive::rotate( char axis, double theta )
{
	int i, j ;
	double** T, **result ;
	T = new double*[4] ; result = new double*[4] ;
	for( i = 0 ; i < 4 ; i++ )
	{
		T[i] = new double[4] ;
		result[i] = new double[4] ;
	}
	
	T[0][3] = T[1][3] = T[2][3] = T[3][0] = T[3][1] = T[3][2] = 0 ;
	T[3][3] = 1 ;
	
	double cos_theta = cos( theta ) ;
	double sin_theta = sin( theta ) ;	
	
	switch( axis )
	{
		case 'x' :
			 T[0][0] = 1 ;
			 T[0][1] = T[0][2] = T[1][0] = T[2][0] = 0 ;
			 T[1][1] = T[2][2] = cos_theta ;
			 T[1][2] = -(T[2][1] = sin_theta) ;
			 break ;
		case 'y' :
			 T[1][1] = 1 ;
			 T[0][1] = T[1][0] = T[1][2] = T[2][1] = 0 ;
			 T[0][0] = T[2][2] = cos_theta ;
			 T[2][0] = -(T[0][2] = sin_theta) ;
			 break ;
		case 'z' :
			 T[2][2] = 1 ;
			 T[0][2] = T[1][2] = T[2][0] = T[2][1] = 0 ;
			 T[0][0] = T[1][1] = cos_theta ;
			 T[0][1] = -(T[1][0] = sin_theta) ;
			 break ;
		default :
			 break ;
	}
	
	if( !isTransformed ) {
		isTransformed = true ;
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		matrix_assign( M, T ) ;
		matrix_assign( N, T, 3 ) ;
	}
	
	else {
		// Multiply M = TM
		matrix_multiply( T, M, result ) ;
		matrix_assign( M, result ) ;
		matrix_multiply( T, N, result, 3 ) ;
	        matrix_assign( N, result, 3 ) ;	
	}
	
	/*
	printf("\nDebug: Transformation matrix = \n") ;
	for( i = 0 ; i < 4 ; i++ ) {
		for( j = 0 ; j < 4 ; j++ ) 
			printf("%d ", M[i][j] ) ;
		printf("\n") ;
	}
	*/
}

Vector3D Primitive::transform_point( const Vector3D p )
{
	double result[3] ;
	int i ;
	for( i=0; i<3; i++ )
	{
		result[i] = M[i][0]*p.x + M[i][1]*p.y + M[i][2]*p.z + M[i][3] ;
	}

	return Vector3D( result[0], result[1], result[2] ) ;
}

Vector3D Primitive::transform_normal( const Vector3D n )
{
	double result[3] ;
	int i ;
	for( i=0; i<3; i++ )
	{
		result[i] = N[i][0]*n.x + N[i][1]*n.y + N[i][2]*n.z ;
	}

	return Vector3D( result[0], result[1], result[2] ) ;
}

// Inverts the transformation matrix M
void Primitive::compute_inverse( void ) 
{
	if( isTransformed ) { 
		// Invoke matrix inversion function
		matrix_invert( M, inverse ) ;
		/*
		printf("\nDebug: Inverse Transformation matrix = \n") ;

		int i, j ;
		for( i = 0 ; i < 4 ; i++ ) {
			for( j = 0 ; j < 4 ; j++ ) printf("%lf ", inverse[i][j]) ;
			printf("\n") ;
		}
		*/
	}
}

Ray Primitive::transform_ray( const Ray& r )
{
	// inverse is assumed to be precomputed - saves time as we do not need to invert the matrix for every ray object intersection
	// Another quick check to save time
	if( !isTransformed ) return r ;

	// Otherwise transform the ray's origin and direction
	double new_origin[3] ;
	int i ;
	for( i=0; i<3; i++ )
	{
		new_origin[i] = 
			inverse[i][0]*r.origin.x + inverse[i][1]*r.origin.y + inverse[i][2]*r.origin.z + inverse[i][3] ;

	}

	double new_direction[3] ;
	for( i=0; i<3; i++ )
	{
		new_direction[i] = 
			inverse[i][0]*r.direction.x + inverse[i][1]*r.direction.y + inverse[i][2]*r.direction.z  ;

	}

	Ray ray = Ray( Vector3D( new_origin[0], new_origin[1], new_origin[2] ), Vector3D( new_direction[0], new_direction[1], new_direction[2] ) ) ;

	// Compute virtual "length" - bit of a pain
	if( r.length == MAX_LENGTH ) ray.length = MAX_LENGTH ;
	else {
		Vector3D point_on_ray( r.origin.x + r.length * r.direction.x, 
				r.origin.y + r.length * r.direction.y,  		
				r.origin.z + r.length * r.direction.z
				);
		Vector3D inverse_point( 
				inverse[0][0] * point_on_ray.x + inverse[0][1] * point_on_ray.y + inverse[0][2] * point_on_ray.z+ inverse[0][3],
				inverse[1][0] * point_on_ray.x + inverse[1][1] * point_on_ray.y + inverse[1][2] * point_on_ray.z+ inverse[1][3],
				inverse[2][0] * point_on_ray.x + inverse[2][1] * point_on_ray.y + inverse[2][2] * point_on_ray.z+ inverse[2][3]					     );	
		ray.length = Vector3D( inverse_point.x - ray.origin.x,
				inverse_point.y - ray.origin.y,
				inverse_point.z - ray.origin.z
				).length() ;
	} // end of else clause	

	//	printf("\nDebug: ray.length = %lf r.length = %lf\n", ray.length, r.length ) ;
	return ray ;
}

Primitive::~Primitive() 
{
	int i ;
	if( isTransformed ) {
		for( i = 0 ; i < 4 ; i++ )
		{
			delete [] M[i] ;
			if( i != 3 ) delete [] N[i] ;
		}

		delete [] M ; delete [] N ;
	}
}

bool Primitive::intersect( Ray& r ) 
{
	// Intersects this object with a ray
	return false ;
}

bool Primitive::shadow_intersect( Ray& r ) 
{
	// Intersects this object with a shadow ray
	return false ;
}

void Primitive::computeBounds( bVolume& b )  { /* Dummy */ } // Compute bounding volume for this primitive

Color Primitive::shade( const Ray& r, int depth )
{
	// Dummy - should not ever be called
	return Color(0,0,0) ;
}

void Primitive::setSurface( Surface* s ) 
{
	surface = s ;
}

// Used for freeing memory and avoiding leak - see simple_sdl.y 's deallocate_mem() function for details
Surface* Primitive::getSurface()
{
	return surface ;
}

// Nothing - just to keep g++ happy :)
vector <CSG_Interval*> Primitive::compute_intervals( const Ray& ray ) 
{
	vector <CSG_Interval*> new_list ;
	return new_list ;
}

// transforms this primitive's bounding box appropriately
void Primitive::transform_box( bVolume& b ) 
{
	// transform each corner of the box
	Vector3D corner[8] ;
	corner[0] = transform_point( Vector3D( b.xmin, b.ymin, b.zmin ) ) ;
	corner[1] = transform_point( Vector3D( b.xmin, b.ymin, b.zmax ) ) ;
	corner[2] = transform_point( Vector3D( b.xmin, b.ymax, b.zmin ) ) ;
	corner[3] = transform_point( Vector3D( b.xmin, b.ymax, b.zmax ) ) ;
	corner[4] = transform_point( Vector3D( b.xmax, b.ymin, b.zmin ) ) ;
	corner[5] = transform_point( Vector3D( b.xmax, b.ymin, b.zmax ) ) ;
	corner[6] = transform_point( Vector3D( b.xmax, b.ymax, b.zmin ) ) ;
	corner[7] = transform_point( Vector3D( b.xmax, b.ymax, b.zmax ) ) ;
	
	// Find min, max of x, y, z for each
	double min_x, min_y, min_z, max_x, max_y, max_z ;
	min_x = max_x = corner[0].x ; 
	min_y = max_y = corner[0].y ; 
	min_z = max_z = corner[0].z ; 

	for( int i = 1 ; i < 8 ; i++ )	
	{
		if( corner[i].x < min_x ) min_x = corner[i].x ;
		else if( corner[i].x > max_x ) max_x = corner[i].x ;
		if( corner[i].y < min_y ) min_y = corner[i].y ;
		else if( corner[i].y > max_y ) max_y = corner[i].y ;
		if( corner[i].z < min_z ) min_z = corner[i].z ;
		else if( corner[i].z > max_z ) max_z = corner[i].z ;
	}

	b.xmin = min_x ;
	b.ymin = min_y ;
	b.zmin = min_z ;
	b.xmax = max_x ;
	b.ymax = max_y ;
	b.zmax = max_z ;
}
