// Abstract graphics primitive class - primitive.h

#ifndef __PRIMITIVE_H
#define __PRIMITIVE_H

#include "vector3D.h"
#include "light.h"
#include "color.h"
#include "ray.h"
#include "object.h"
#include "surface.h"
#include "bvolume.h"

#include <vector>

using std::vector ;

// Forward declaration
class CSG_Interval ;

// This is a class that can serve as an interface for any primitive object, be it a sphere or a quadric
// It is not an abstract class as it contains some attributes of the primitive:
// Namely, the transformation matrix to be applied and the surface properties for the primitive
class Primitive : public Object
{
	public:
	// Surface properties for the primitive
	Surface* surface ;
	
	// Transformation matrix to apply (Using Foley's premultiplication convention)
	// Notice that only affine transformations are permitted as of now - this for simplicity
	double** M ; // Proper tranformation matrix
	double** N ; // Normal transformation matrix
	double inverse[4][4] ; // Inverse transformation for rays
	bool isTransformed ; // Indicates if M is to be used or not - mainly to avoid doubleing point error

	Primitive() { isTransformed = false ; } // Default constructor
	Primitive( const Primitive& p ) ; // Copy constructor
	Primitive( Surface* s ) ;  // Initializes a primitive with the given surface properties
	
	// Some functions that will be used to apply transformations to this primitive
	virtual void translate( double X, double Y, double Z ) ; // Translates this object
	virtual void scale( double Sx, double Sy, double Sz ) ; // Scales about the origin
	virtual void rotate( char axis, double theta ) ; // Rotates by theta about axis 'X', 'Y' or 'Z'
	virtual void compute_inverse( void ) ; // Freezes the object - basically computes inverse of transformation matrix	
	virtual void setSurface( Surface* s ) ; // Set surface properties as specified in s
	virtual Surface* getSurface() ; // Get surface of this primitive
	// Some virtual functions that will be late bound as usual :) - These override the pure virtual functions that are declared in abstract class Object and will further be overridden by the actual primitive
	virtual bool intersect( Ray& r ) ; // Intersects this primitive with a ray
	virtual bool shadow_intersect( Ray& r ) ; // Intersects this primitive with a shadow ray
	virtual void computeBounds( bVolume& b ) ; // Compute bounding volume for this primitive
	virtual Color shade( const Ray& r, int depth ) ; // Shades this object by passing parameters to the surface shader
	
	// Some transformation methods common to all primitives - see if they are needed later
	Ray transform_ray( const Ray& r ) ; // Transforms ray in inverse if necessary and returns result ray	
	Vector3D transform_point( const Vector3D p ) ; // Transforms this point forward
	Vector3D transform_normal( const Vector3D normal ) ; // Transforms the normal appropriately
	void transform_box( bVolume& b ) ; // Transforms this object's bounding box appropriately	
	// virtual method to compute interval list of ray-primitive intersection
	virtual vector <CSG_Interval*> compute_intervals( const Ray& ) ;
	virtual ~Primitive() ; // Virtual Destructor
};

#endif
