// Ray class method definitions - ray.cpp

#include "object.h"
#include "ray.h"
#include "bvolume.h"

#include <queue>
#include <map>
#include <vector>
#include <assert.h> // For assert()

using std::queue ;
using std::priority_queue ;
using std::deque ;
using std::map ;
using std::vector ;

int MAX_RAYTRACE_DEPTH = 10 ; // The maximum depth to ray trace recursively by default
extern bVolumeHierarchy* boxTree ; // Precomputed bounding volume hierarchy
extern unsigned long global_object_count ; // For debugging purposes
extern map<Object*,int> objectMap ; // For debugging purposes
extern vector<Object*> infiniteObjectList ; // The list of infinite objects like planes

// Creates a ray from the eye along the direction specified by the 3D vector dir
Ray::Ray(Vector3D eye, Vector3D dir) 
{
	origin = eye ;
	direction = Vector3D::normalize(dir) ;
	length = MAX_LENGTH ; // Initialize to infinity
	object = NULL ; // Initialize

	// By default, ray travels in vacuum and starts outside - so nd ( nesting depth ) = 0
	n = 1 ;
	inside = false ;
}

/*
// Performs ray-object intersections for all the objects in the scene
// From this we determine the primary object (if any) intersecting the ray
bool Ray::trace( const vector<Object*>& objects) 
{
	// Bug in old code !
	// length = MAX_LENGTH ;
	object = NULL ;

	// Dynamic binding - appropriate virtual function will be called ( hopefully ) 
	for( vector<Object*>::const_iterator it = objects.begin() ; it != objects.end() ; it++ ) {
		(*it) -> intersect( *this ) ;	
	}
	return (object != NULL);
}
*/

/* New ray trace code when bounding box optimizations are done */
/* Uses the Kay-kajiya algorithm */

/* The actual ray tracing method optimized using bounding volume hierarchies 
 */
bool Ray::trace()
{
	// The important idea is to first order all the bounding boxes by distance along the ray
	// This distance computation is accomplished by a call to the ray-bounding box or ray-primitive
	// intersection routine just before insertion into the heap
	// Thereafter, the comparison result is memoized - we save calls to intersect() when building 
	// and manipulating the heap

	// The algorithm ( Kay-Kajiya '86 ) is as follows:
	// Insert( Heap, Box/Primitive b ) : 
	// 1 Execute b -> intersect( ray ) whether primitive or box
	// 2 If result successful insert b into heap
	// 
	// Main algorithm
	// ---------------
	// 1 Insert root node of bounding volume hierarchy into heap
	// 2 while( heap is not empty ) do
	// 3 	extract nearest node from the heap
	// 4	if the nearest node is a primitive (leaf)
	// 5 		then ray.object = this primitive ; return ;
	// 6	else 
	// 7		nearest node is a bounding box, say B
	// 8		determine the children of B in the hierarchy
	// 9		insert children into the heap
	// 10 end do

	static priority_queue< bVolumeNode*, deque<bVolumeNode*>, bVolumeNodeCompare > heap ;
	bool ray_intersects_box ;

	// Clear the heap
	while( !heap.empty() ) heap.pop() ;
	
	bVolumeNode* newPlaneNode = new bVolumeNode[ infiniteObjectList.size() ] ;
	// First load all infinite objects like planes into the heap
	for( int i = 0 ; i < infiniteObjectList.size() ; i++ )
	{
			((newPlaneNode[i]).box).obj = infiniteObjectList[i] ;
			(newPlaneNode[i]).nc = 0 ;
			(newPlaneNode[i]).distance_from_ray = MAX_LENGTH ;
			ray_intersects_box = (newPlaneNode[i]).intersect( *this ) ; // Calls appropriate algorithm 
			if( ray_intersects_box )
			{
					//printf("\nDebug: Plane number %d intersects ray at distance %lf\n", i, newPlaneNode -> distance_from_ray ) ;
					heap.push( newPlaneNode + i ) ;
			}
	}
	
	// Now for the finite objects
	if( boxTree -> root ) 
	{	
		( boxTree -> root ) -> distance_from_ray = MAX_LENGTH ;
		ray_intersects_box = ( boxTree -> root ) -> intersect( *this ) ; // Calls appropriate algorithm 
		if( ray_intersects_box )
			heap.push( boxTree -> root ) ;
		else {
			( ( boxTree -> root ) -> distance_from_ray ) = MAX_LENGTH ;
			//printf("\nDebug: No root\n") ;
		}
	}
	
	/* Do not forget to add code before this while loop to insert infinite primitives like planes
	 * into the initial heap along with the root bounding box - TODO
	 */
	while( !heap.empty() ) {
		bVolumeNode* nearestNode = heap.top() ; heap.pop() ;
		// Test if the nearest node is a leaf
		int numchildren ;
		numchildren = nearestNode -> nc ;
		if( !( numchildren ) ) {
			global_object_count++ ;
			//printf("\nDebug: Leaf intersects ray at distance %lf\n", nearestNode -> distance_from_ray ) ;
			//assert( ( nearestNode -> box ).obj ) ;
			//assert( ( nearestNode -> box ).obj == this -> object ) ;
			nearestNode -> distance_from_ray = MAX_LENGTH ;
			delete [] newPlaneNode ;
			return true ;
		}

		else {
			//printf("\nDebug: Traversing Internal node %p at a distance of %lf with %d children\n", nearestNode, nearestNode -> distance_from_ray, nearestNode -> nc ) ;
			// Not a leaf but a bounding box
			// This must have children ; push them all into the heap after ray tracing them
			//printf("\nDebug: numchildren = %d\n", numchildren ) ;
			for( int i = 0 ; i < numchildren ; i++ ) {
				bVolumeNode* child = ( nearestNode -> childList )[i] ;
				child -> distance_from_ray = MAX_LENGTH ;
				//printf("\nDebug: childList[%d] = %p\n", i, (child -> box).obj ) ;
				ray_intersects_box = child -> intersect( *this ) ;
				if( ray_intersects_box )
				{
					//printf("\nDebug: Pushing internal node child %p of internal node with distance_from_ray = %lf\n", child, (child -> distance_from_ray) ) ;
					heap.push( child ) ;
				}
				else child -> distance_from_ray = MAX_LENGTH ;
			}	// end of for
			nearestNode -> distance_from_ray = MAX_LENGTH ;
		}	// end of else
	} 	// end of while
	
	// No primitives intersect inside the loop. So return false.
	delete [] newPlaneNode ;
	return false ;
}

/* A slightly more optimized method for shadow rays
 * Uses shadow caching for maximum benefit
 * The idea is to cache an object once it casts a shadow and test it first the next time
 * Leads to cache hits with high probability and saves quite some time
 */
bool Ray::shadow_trace()
{
	static Object* shadowCache = NULL ;
	static queue<bVolumeNode*> heap ; // We do not need a priority queue for shadow rays - it is wasteful
	bool ray_intersects_box ;
	
	// First test the shadow cache if available
	if( shadowCache )
	{
		if( shadowCache -> shadow_intersect( *this ) ) /* Cache hit */ return true ;
		else /* Invalidate cache - Cache miss */ shadowCache = NULL ;
	}

	// Clear the heap
	while( !heap.empty() ) heap.pop() ;

	if( boxTree -> root ) 
	{
		( boxTree -> root ) -> distance_from_ray = MAX_LENGTH ;
		ray_intersects_box = ( boxTree -> root ) -> shadow_intersect( *this ) ; // Calls appropriate algorithm 
		if( ray_intersects_box )
			heap.push( boxTree -> root ) ;
		else 
			return false ;
	}

	/* Do not forget to add code before this while loop to insert infinite primitives like planes
	 * into the initial heap along with the root bounding box - TODO
	 */
	while( !heap.empty() ) {
		bVolumeNode* nearestNode = heap.front() ; heap.pop() ;
		// Test if the nearest node is a leaf
		int numchildren ;
		numchildren = nearestNode -> nc ;
		if( !( numchildren ) ) {
			shadowCache = ( nearestNode -> box ).obj ;
			nearestNode -> distance_from_ray = MAX_LENGTH ;
			return true ;
		}

		else {
			// Not a leaf but a bounding box
			// This must have children ; push them all into the heap after ray tracing them
			for( int i = 0 ; i < numchildren ; i++ ) {
				bVolumeNode* child = ( nearestNode -> childList )[i] ;
				ray_intersects_box = child -> shadow_intersect( *this ) ;
				if( ray_intersects_box )
					heap.push( child ) ;
				else child -> distance_from_ray = MAX_LENGTH ;
			}	// end of for
			nearestNode -> distance_from_ray = MAX_LENGTH ;
		}	// end of else
	} 	// end of while
	
	// No primitives intersect inside the loop. So return false.
	return false ;
}

/*
// Performs shadow ray-object intersections for all the objects in the scene
// For optimization - unnecessary to find nearest hit for shadow rays
bool Ray::shadow_trace( const vector<Object*>& objects) 
{
	// length = MAX_LENGTH ;

	// Dynamic binding - appropriate virtual function will be called ( hopefully ) 
	for( vector<Object*>::const_iterator it = objects.begin() ; it != objects.end() ; it++ ) {
		bool hit = (*it) -> shadow_intersect( *this ) ;
		if( hit ) return true ;	
	}

	return false ;
}
*/

// Simply a wrapper for the primary object's shader method
Color Ray::shade( int depth ) {
	Color result( 0.0, 0.0, 0.0 ) ;
	if( depth > MAX_RAYTRACE_DEPTH ) return result ;	
	if( object == NULL ) ; // No meaning really - just a safety check
	else {
		result = (*object).shade( *this , depth ) ; // Dynamic binding - polymorphism
		//printf("\nDebug: Shade color = (%lf,%lf,%lf)\n", result.red, result.green, result.blue) ;
	}
	return result ;
}
