/* simple_sdl.l - lex file that tokenizes a simple input scene description file 
 */

%{

#include <iostream>
#include <stdlib.h> // For atof()

#include "vector3D.h" // For Vector3D type
#include "color.h" // For Color type
#include "light.h" // For light types
#include "wrappers.h" // For C wrapper struct types
#include "sdl.h" // For yacc definitions

using std::cerr ;

void yyerror( char* str ) ;

%}

DIGIT [0-9]
DIGITS {DIGIT}+
OPTIONAL_FRACTION \.{DIGITS}
SIGN [+-]
NUMBER {SIGN}?{DIGITS}{OPTIONAL_FRACTION}?

WHITESPACE [ \t\n]+
%option nomain
%option yylineno

%%

"/*"        {
	
	// Multiline comment
	register int c;

	for ( ; ; )
	{
		while ( (c = yyinput()) != '*' &&
				c != EOF )
			;    /* eat up text of comment */

		if ( c == '*' )
		{
			while ( (c = yyinput()) == '*' )
				;
			if ( c == '/' )
				break;    /* found the end */
		}

		if ( c == EOF )
		{
			yyerror( "EOF in comment" );
			break;
		}
	}
}

\".*\" {
	/* Match all strings (hopefully) */
	char* result = strdup( yytext ) ;
	int i ;
	for( i = 1 ; yytext[i+1] ; i++ ) {
		result[i-1] = yytext[i] ;
	}
	result[i-1] = '\0' ;
	//printf("\nDebug: result = %s", result ) ;
	yylval.filename = result ;
	return __STRING_TOKEN ;	
}

eye {
	return __EYE_TOKEN ;	
}

lookat {
	return __LOOKAT_TOKEN ;
}

up {
	return __UP_TOKEN ;
}

fov {	
	return __FOV_TOKEN ;
}

background {
	return __BACKGROUND_TOKEN ;
}

maxdepth {
	return __MAXDEPTH_TOKEN ;
}

height {
	return __HEIGHT_TOKEN ;
}

width {
	return __WIDTH_TOKEN ;
}

eye_candy {
	return __EYE_CANDY_TOKEN ;
}

false {
	return __FALSE_TOKEN ;
}

true {
	return __TRUE_TOKEN ;
}

light {
	return __LIGHT_TOKEN ;
}

ambient {
	return __AMBIENT_TOKEN ;
}

directional {
	return __DIRECTIONAL_TOKEN ;
}

point {
	return __POINT_TOKEN ;
}

object {
	return __OBJECT_TOKEN ;
}

translate {
	return __TRANSLATE_TOKEN ;
}

scale {
	return __SCALE_TOKEN ;
}

rotate {
	return __ROTATE_TOKEN ;
}

sphere {
	return __SPHERE_TOKEN ;
}

plane {
	return __PLANE_TOKEN ;
}

box {
	return __BOX_TOKEN ;
}

triangle {
	return __TRIANGLE_TOKEN ;
}

import {
	return __IMPORT_TOKEN ;
}

from {
	return __FROM_TOKEN ;
}

surface {
	return __SURFACE_TOKEN ;
}

texture {
	return __TEXTURE_TOKEN ;
}

filter {
	return __FILTER_TOKEN ;
}

bilinear {
	return __BILINEAR_TOKEN ;
}

union {
	return __UNION_TOKEN ;
}

intersection {
	return __INTERSECTION_TOKEN ;
}

difference {
	return __DIFFERENCE_TOKEN ; 
}

area {
	return __AREA_TOKEN ;
}

xminus {
	return __XMINUS_TOKEN ;
}

yminus {
	return __YMINUS_TOKEN ;
}

zminus {
	return __ZMINUS_TOKEN ;
}

xplus {
	return __XPLUS_TOKEN ;
}

yplus {
	return __YPLUS_TOKEN ;
}

zplus {
	return __ZPLUS_TOKEN ;
}

path {
	return __PATH_TOKEN ;
}

schlick {
	return __SCHLICK_TOKEN ;
}

{NUMBER} {
	yylval.value = atof(yytext) ;
	return __NUMBER_TOKEN ;
		 }

\{ {
	return __OPEN_BRACE_TOKEN ;
}

\} {
	return __CLOSE_BRACE_TOKEN ;
}

x {
	return __X_TOKEN ;
}

y {
	return __Y_TOKEN ;
}

z {
	return __Z_TOKEN ;
}

{WHITESPACE} {

			 }

. { 

	cerr << "Line " << yylineno << " - Unrecognized symbol " << yytext << "\n" ;
}

%%

int yywrap() {
	return 1 ;
}

void yyerror( char* str ) {
	cerr << "Error when parsing scene description file at line " << yylineno << " - " << str << "\n" ;
	exit(0) ;
}
