/* simple_sdl.y - Yacc file that parses a simple input scene descripion file
 * To be used in conjunction with simple_sdl.l
 */

%token __EYE_TOKEN __LOOKAT_TOKEN __FOV_TOKEN __UP_TOKEN __BACKGROUND_TOKEN __LIGHT_TOKEN __AMBIENT_TOKEN __DIRECTIONAL_TOKEN __POINT_TOKEN __OBJECT_TOKEN __TRANSLATE_TOKEN __SCALE_TOKEN __ROTATE_TOKEN __SPHERE_TOKEN __PLANE_TOKEN __SURFACE_TOKEN __NUMBER_TOKEN __X_TOKEN __Y_TOKEN __Z_TOKEN __OPEN_BRACE_TOKEN __CLOSE_BRACE_TOKEN __UNION_TOKEN __INTERSECTION_TOKEN __DIFFERENCE_TOKEN __MAXDEPTH_TOKEN __EYE_CANDY_TOKEN __TRUE_TOKEN __FALSE_TOKEN __BOX_TOKEN __HEIGHT_TOKEN __WIDTH_TOKEN __TRIANGLE_TOKEN __IMPORT_TOKEN __FROM_TOKEN __STRING_TOKEN __TEXTURE_TOKEN __FILTER_TOKEN __BILINEAR_TOKEN __AREA_TOKEN __XMINUS_TOKEN __YMINUS_TOKEN __ZMINUS_TOKEN __XPLUS_TOKEN __YPLUS_TOKEN __ZPLUS_TOKEN __PATH_TOKEN __SCHLICK_TOKEN

%token <value> __NUMBER_TOKEN
%token <filename> __STRING_TOKEN

%{
	#include <stdio.h> // For FILE
	#include "vector3D.h" // For Vector3D type
	#include "color.h" // For Color type
	#include "light.h" // For light types
	#include "object.h" // For Object 
	#include "surface.h" // For Surface
	#include "primitive.h" // For Primitive
	#include "sphere.h" // For Sphere
	#include "plane.h" // For Plane
	#include "box.h" // For Box
	#include "arealight.h" // For Area lights
	#include "triangle.h" // For Triangle
	#include "csg.h" // For CSG
	#include "wrappers.h" // Some wrapper structures
	#include "ray.h" // For setting max ray tracing depth
	#include "texture.h" // For texture mapping methods/definitions

	#include <vector>	// For STL vector
	#include <map>		// For STL Map
	#include <math.h> // For PI

	using std::vector ;
	using std::map ;

	/* Some global declarations */
	extern int yylex(void);  
	extern void yyerror(char*) ;

	extern FILE* yyin ; /* Defined by lex */

/* Important extern declarations from init.cpp */
	extern FILE* inputFile ;
	extern Vector3D eye, lookat, up, Du, Dv, Vp ;
	extern double fov ;
	extern Color background ;
	extern vector<Light*> lightList ;
	extern vector<Object*> objectList ;
	extern vector<Object*> infiniteObjectList ;
	extern vector<AreaLight*> areaLightList ;
	extern map<Object*,int> objectMap ;
	extern Surface* currentSurface ;

	extern int MAX_RAYTRACE_DEPTH ;
	extern bool eye_candy ;
	extern int width, height ;
	extern int global_texture_filter_setting ;
	extern bool enable_path ;
	extern int paths_per_pixel ;

/* Extern method definitions from init.c to be called from main() */
	extern void parse_command_line( int argc, char* argv[] ) ;
	extern void init_defaults() ;
	extern void init_tracer() ; 
	extern void traceAll() ; 
	extern void pathTraceAll() ;
	extern void write_image() ;
	%}
	
	%type <lorient> orientation
	%type <vec> coordinate
	%type <c> color
	%type <ltype> light_type
	%type <trinfo> transformation
	%type <axis_name> axis
	%type <object_marker> object_statement_list
	%type <isPrimitive> object_declaration object_description object_statement

%union{
		struct __wrap_Vector3D vec ;
		struct __wrap_Color c ;
		struct __wrap_transformation trinfo ;
		double value ;
		char* filename ;
		int ltype ;
		char axis_name ;
		struct __wrap_marker object_marker ;
		bool isPrimitive ;
		int lorient ;
	}

%%

Scene: statements
	 ;

statements: statements statement
		  |
;

statement: 
eye_statement
| lookat_statement
| up_statement
| fov_statement
| background_statement
| maxdepth_statement
| eye_candy_statement
| image_statement
| light_statement
| surface_statement { ; }
| object_statement { ; }
| filter_statement
| path_statement
;

coordinate: __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN {
		  $$.x = $1 ; $$.y = $2 ; $$.z = $3 ;
		}
;

color: __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN {
	 $$.red = $1 ; $$.green = $2 ; $$.blue = $3 ;
	   }
;

eye_statement: __EYE_TOKEN coordinate {
			 /* Set eye coordinates */
			   eye = Vector3D($2) ;
		   }
;

lookat_statement:  __LOOKAT_TOKEN coordinate {
				/* Set lookat vector */
			  lookat = Vector3D($2) ;
		  }
;

fov_statement: __FOV_TOKEN __NUMBER_TOKEN {
			 /* Set field of view */
			   fov = $2 ;
		   }
;

up_statement: __UP_TOKEN coordinate {
			/* Set view up vector */
			  up = Vector3D($2) ;
		  }
;

background_statement: __BACKGROUND_TOKEN color {
					/* Set background color */
				  background = Color($2) ;
			  }
;
path_statement: __PATH_TOKEN __NUMBER_TOKEN {
	      	enable_path = true ; paths_per_pixel = (int)$2 ;
}
;

maxdepth_statement: __MAXDEPTH_TOKEN __NUMBER_TOKEN {
				  MAX_RAYTRACE_DEPTH = (int)$2 ;
}
;

eye_candy_statement: __EYE_CANDY_TOKEN __TRUE_TOKEN {
				   eye_candy = true ;
}
| __EYE_CANDY_TOKEN __FALSE_TOKEN {
		eye_candy = false ;
}
;

filter_statement: __FILTER_TOKEN __BILINEAR_TOKEN {
				global_texture_filter_setting = BILINEAR_FILTER ;
}
;

image_statement: __HEIGHT_TOKEN __NUMBER_TOKEN {
			   height = (int)$2 ;
}
| __WIDTH_TOKEN __NUMBER_TOKEN {
	width = (int)$2 ;
}
;

orientation: __XMINUS_TOKEN { $$ = NEGATIVE_X ; }
| __YMINUS_TOKEN { $$ = NEGATIVE_Y ; }
| __ZMINUS_TOKEN { $$ = NEGATIVE_Z ; }
| __XPLUS_TOKEN  { $$ = POSITIVE_X ; }
| __YPLUS_TOKEN { $$ = POSITIVE_Y ; }
| __ZPLUS_TOKEN  { $$ = POSITIVE_Z ; }
;

/* Add more light types later */
light_type: __AMBIENT_TOKEN { $$ = AMBIENT ; }
		  | __POINT_TOKEN { $$ = POINT ; }
| __DIRECTIONAL_TOKEN { $$ = DIRECTIONAL ; }
| __AREA_TOKEN { $$ = AREA ; }
;

light_statement: __LIGHT_TOKEN color light_type {
			 /* Create a new ambient light source */
			 AmbientLight* new_light = new AmbientLight( $2.red, $2.green, $2.blue ) ;
			 lightList.push_back( new_light ) ;
		 }
| __LIGHT_TOKEN color light_type coordinate {
			 /* Check type */
			 if( $3 == POINT ) {
			 	PointLight* new_light = new PointLight( Vector3D($4), $2.red, $2.green, $2.blue ) ;
			 	lightList.push_back( new_light ) ;
			 }
			 else if( $3 == DIRECTIONAL ) {
			 	DirectionalLight* new_light = new DirectionalLight( Vector3D($4), $2.red, $2.green, $2.blue ) ;
			 	lightList.push_back( new_light ) ;
			 }
		}
| __LIGHT_TOKEN color light_type orientation __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN {
	// This is an area light
	EmissiveSurface* emissive = new EmissiveSurface( $2 ) ;
	AreaLight* newlight = new AreaLight( emissive, $4, $5, $6, $7, $8, $9 ) ;
	objectList.push_back( newlight ) ;
	areaLightList.push_back( newlight ) ;
}	
;

surface_statement: __SURFACE_TOKEN color __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN {
			 /* Change the surface modal attribute to a phong surface */
			 currentSurface = new PhongSurface( Color($2), $3, $4, $5, $6, $7, $8, $9 ) ;
		   }
| __SURFACE_TOKEN __OPEN_BRACE_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __TEXTURE_TOKEN __STRING_TOKEN __CLOSE_BRACE_TOKEN {
			/* Texture mapped phong surfaces */
			currentSurface = new PhongSurface( Color(0, 0, 0), $3, $4, $5, $6, $7, $8, $9 ) ;
			Texture* newTexture = new Texture() ;
			newTexture -> loadFromPPM( $11 ) ;
			currentSurface -> setTexture( newTexture ) ;
			}
| __SURFACE_TOKEN __SCHLICK_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN __NUMBER_TOKEN
{
	/* Schlick surface */
	currentSurface = new SchlickSurface( $3, $4, $5, $6, $7 ) ;
}
;

/* Add more objects later */
object_statement: __OBJECT_TOKEN __OPEN_BRACE_TOKEN object_description __CLOSE_BRACE_TOKEN {
				$$ = $3 ;
		  }
;

object_description: object_declaration transformation_sequence surface_statement {
		  // Set surface properly
		  (objectList.back()) -> setSurface( currentSurface ) ;
		  // Freeze object by computing matrix inverse
		  (objectList.back()) -> compute_inverse() ;
		  $$ = $1 ; 		
}	
| object_declaration transformation_sequence {
		  // Freeze object by computing matrix inverse
		  if( objectList.size() != 0 )
		  	(objectList.back()) -> compute_inverse() ;
		  $$ = $1 ;
}
;

transformation_sequence: transformation_sequence transformation {
	/*
	// Apply this transformation correctly to the last object in objectList
	// This will be complicated to parse when we have CFG - for now manage

// Get transformation type, parameters
	*/
	switch( $2.type ) {
		case TRANSLATE_TRANSFORMATION:
		( objectList.back() ) -> translate( $2.x, $2.y, $2.z ) ;
		break ;

case SCALE_TRANSFORMATION:
		( objectList.back() ) -> scale( $2.x, $2.y, $2.z ) ;
		break ;

case ROTATE_TRANSFORMATION:
		( objectList.back() ) -> rotate( $2.axis, $2.theta ) ;
		break ;

default:
	   break ;
	}
}				
|
;

transformation: __TRANSLATE_TOKEN coordinate {
			  $$.type = TRANSLATE_TRANSFORMATION ;
	$$.x = $2.x ;
	$$.y = $2.y ;
	$$.z = $2.z ;
}	
| __SCALE_TOKEN coordinate {
	$$.type = SCALE_TRANSFORMATION ;
	$$.x = $2.x ;
	$$.y = $2.y ;
	$$.z = $2.z ;
}
| __ROTATE_TOKEN axis __NUMBER_TOKEN {
	$$.type = ROTATE_TRANSFORMATION ;
	$$.axis = $2 ;
	$$.theta = $3 * M_PI / 180 ;
}
;

axis: __X_TOKEN {
	$$ = 'x' ;
}
| __Y_TOKEN {
	$$ = 'y' ;
}
| __Z_TOKEN {
	$$ = 'z' ;
}
;

object_declaration: primitive_declaration {
				  $$ = true ;
}
| CSG_declaration {
	$$ = false ;
}
;

primitive_declaration: __SPHERE_TOKEN coordinate __NUMBER_TOKEN {
					 Sphere* new_sphere = new Sphere( currentSurface, Vector3D($2), $3 ) ;
			  objectList.push_back( new_sphere ) ;
		  }
| __PLANE_TOKEN coordinate coordinate {
	Plane* new_plane = new Plane( currentSurface, Vector3D( $2 ), Vector3D( $3 ) ) ; 
	infiniteObjectList.push_back( new_plane ) ;
	}
| __BOX_TOKEN coordinate coordinate {
	Box* new_box = new Box( currentSurface, Vector3D( $2 ), Vector3D( $3 ) ) ;
	objectList.push_back( new_box ) ;
}
| __TRIANGLE_TOKEN coordinate coordinate coordinate {
	// Flat shaded triangle
	Triangle* new_triangle = new Triangle( currentSurface, Vector3D($2), Vector3D($3), Vector3D($4) ) ;
	objectList.push_back( new_triangle ) ;
}
| __TRIANGLE_TOKEN coordinate coordinate coordinate coordinate coordinate coordinate {
	// Phong shaded triangle
	Triangle* new_triangle = new Triangle( currentSurface, Vector3D($2), Vector3D($4), Vector3D($6), Vector3D($3), Vector3D($5), Vector3D($7) ) ;
	objectList.push_back( new_triangle ) ;
}
| __IMPORT_TOKEN __FROM_TOKEN __STRING_TOKEN {
	// Import object from triangle file
	printf("\nReading object from file...\n") ;
	FILE* fp = fopen( $3, "r" ) ;

if( !fp ) {
		printf("\nError: File not found: %s\n", $3 ) ;
	}

else {
		Triangle* new_triangle ;
		char buf[10] ; // Dummy buffer
		// Read triangles from the file in rayshade's format (all phong shaded)
		fscanf( fp, "%s", buf ) ; 
		while( !feof(fp) ) {

//printf("\nDebug: buf = %s\n", buf ) ;

Vector3D v1, v2, v3, n1, n2, n3 ;
			double read ;

fscanf(fp, "%lf", &read ) ; v1.x = read ;
			fscanf(fp, "%lf", &read ) ; v1.y = read ;
			fscanf(fp, "%lf", &read ) ; v1.z = read ;
			fscanf(fp, "%lf", &read ) ; n1.x = read ;
			fscanf(fp, "%lf", &read ) ; n1.y = read ;
			fscanf(fp, "%lf", &read ) ; n1.z = read ;
			fscanf(fp, "%lf", &read ) ; v2.x = read ;
			fscanf(fp, "%lf", &read ) ; v2.y = read ;
			fscanf(fp, "%lf", &read ) ; v2.z = read ;
			fscanf(fp, "%lf", &read ) ; n2.x = read ;
			fscanf(fp, "%lf", &read ) ; n2.y = read ;
			fscanf(fp, "%lf", &read ) ; n2.z = read ;
			fscanf(fp, "%lf", &read ) ; v3.x = read ;
			fscanf(fp, "%lf", &read ) ; v3.y = read ;
			fscanf(fp, "%lf", &read ) ; v3.z = read ;
			fscanf(fp, "%lf", &read ) ; n3.x = read ;
			fscanf(fp, "%lf", &read ) ; n3.y = read ;
			fscanf(fp, "%lf", &read ) ; n3.z = read ;

//printf("\nDebug: Passed double read\n") ;
			new_triangle = new Triangle( currentSurface, v1, v2, v3, n1, n2, n3 ) ;	
			objectList.push_back( new_triangle ) ;
			objectMap[ new_triangle ] = objectList.size() ;
			fscanf( fp, "%s", buf ) ; 
		} // end while
	} // end else
}
;

object_statement_list: object_statement_list object_statement {
					 $$.isPrimitive[ $$.length ] = $2 ;
	$$.length++ ;
}  
| object_statement {
	/*
	// This is the first element in this list
	// Place marker in objectList
	*/
	$$.start = objectList.size() - 1 ;
	$$.isPrimitive[0] = $1 ;
	$$.length++ ;
}
;

CSG_declaration: __UNION_TOKEN __OPEN_BRACE_TOKEN object_statement_list __CLOSE_BRACE_TOKEN {
			   /* The invariant maintained is that the objects whose union is to be performed now occupy contiguous list positions 
	// Starting from the object marker corresponding to the object_statement_list upto the end of the list
	// Just build up the union and push the result into the list after removing the previous objects
	*/
	CSG_Object* new_object = new CSG_Object() ;
	vector <CSG_Object*> union_list ;
	int i, k = 0 ;
	for( i = $3.start ; i < objectList.size() ; i++ )
	{
		if( $3.isPrimitive[k] ) {
			CSG_Object* newleaf = new CSG_Object( static_cast <Primitive*> (objectList[i]) ) ;
			union_list.push_back( newleaf ) ;
		}
		else
			union_list.push_back( static_cast <CSG_Object*> (objectList[i]) ) ;
		k++ ;
	}

new_object -> buildAsUnion( union_list ) ;
	objectList.erase( objectList.begin() + $3.start, objectList.begin() + objectList.size() ) ;
	objectList.push_back( new_object ) ;
}	
| __INTERSECTION_TOKEN __OPEN_BRACE_TOKEN object_statement_list __CLOSE_BRACE_TOKEN {
	/* The invariant mainta TODO ined is that the objects whose union is to be performed now occupy contiguous list positions 
	// Starting from the object marker corresponding to the object_statement_list upto the end of the list
	// Just build up the intersection and push the result into the list after removing the previous objects
	*/
	CSG_Object* new_object = new CSG_Object() ;
	vector <CSG_Object*> union_list ;
	int i, k = 0 ;
	for( i = $3.start ; i < objectList.size() ; i++ )
	{
		if( $3.isPrimitive[k] ) {
			CSG_Object* newleaf = new CSG_Object( static_cast <Primitive*> (objectList[i]) ) ;
			union_list.push_back( newleaf ) ;
		}
		else
			union_list.push_back( static_cast <CSG_Object*> (objectList[i]) ) ;
		k++ ;
	}

new_object -> buildAsIntersection( union_list ) ;
	objectList.erase( objectList.begin() + $3.start, objectList.begin() + objectList.size() ) ;
	objectList.push_back( new_object ) ;
}	
| __DIFFERENCE_TOKEN __OPEN_BRACE_TOKEN object_statement_list __CLOSE_BRACE_TOKEN {
	/* The invariant mainta TODO ined is that the objects whose union is to be performed now occupy contiguous list positions 
	// Starting from the object marker corresponding to the object_statement_list upto the end of the list
	// Just build up the difference and push the result into the list after removing the previous objects
	*/
	CSG_Object* new_object = new CSG_Object() ;
	vector <CSG_Object*> union_list ;
	int i, k = 0 ;
	for( i = $3.start ; i < objectList.size() ; i++ )
	{
		if( $3.isPrimitive[k] ) {
			CSG_Object* newleaf = new CSG_Object( static_cast <Primitive*> (objectList[i]) ) ;
			union_list.push_back( newleaf ) ;
		}
		else
			union_list.push_back( static_cast <CSG_Object*> (objectList[i]) ) ;
		k++ ;
	}

	new_object -> buildAsDifference( union_list ) ;
	objectList.erase( objectList.begin() + $3.start, objectList.begin() + objectList.size() ) ;
	objectList.push_back( new_object ) ;
}	
;

%%

// Deallocates all the memory allocated for lights and objects
void deallocate_mem( void )
{
	for( int i = 0 ; i < lightList.size() ; i++ )
		delete lightList[i] ;
	for( int i = 0 ; i < objectList.size() ; i++ )
		delete objectList[i] ;
}
	
/* The main program at last :) */
int main( int argc, char* argv[] ) 
{
	init_defaults() ;
	parse_command_line( argc, argv ) ;
	yyin = inputFile ;
	yyparse() ;
	init_tracer() ;
	if ( enable_path ) pathTraceAll() ;
	else traceAll() ;
	write_image() ;
	printf("Done.\n") ;
	deallocate_mem() ;
	return 0 ;
}

