// sphere.cpp - Implements intersection and shading methods for a sphere

#include "sphere.h"
#include "csg.h"
#include "vector3D.h"

#include <math.h>

// Constructor
Sphere::Sphere( Surface* s, const Vector3D& c, double r ) : Primitive(s)
{
	centre = c ;
	radius = r ;
	radius_sq = r * r ; // Precompute r*r - it will be useful
}

// Copy constructor
Sphere::Sphere( const Sphere& s ) : Primitive()
{
	int i, j ;
	surface = s.surface ;
	centre = s.centre ;
	radius = s.radius ;
	radius_sq = s.radius_sq ;
	isTransformed = s.isTransformed ;
	if( s.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = s.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = s.N[i][j] ;
			}
	}
}

// First a faster, quicker method just to determine if a shadow ray hits this guy
bool Sphere::shadow_intersect( Ray& r )
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;

	double CLx = centre.x - ray.origin.x ;
	double CLy = centre.y - ray.origin.y ;
	double CLz = centre.z - ray.origin.z ;

	double CLR = ray.direction.dot( CLx, CLy, CLz ) ;
	if( CLR < 0 ) return false ;
	if( radius_sq + (CLR*CLR) - (CLx*CLx) - (CLy*CLy) - (CLz*CLz) < 0 )
		return false ;
	else return true ;
}

// The crucial intersect() method - implements ray object intersection method for a sphere
// Determines if this sphere is the first along the line of sight from the eye to intersect the given ray
// If not, nothing is done otherwise the ray's object is updated to this sphere, which could of course again 
// be overridden by another, closer object.
// We use some tricks for early rejection to avoid computation - more optimizations will come later.
// Finally, this code is modified for when rays can originate inside the sphere (after refraction is implemented)
// We use the boolean variable inside to indicate if the ray originated 'inside' of a closed object
// This way, it becomes easier and faster to do the intersection calculation
bool Sphere::intersect( Ray& r )
{
	Ray ray ;
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;
	
	// Compute direction ratios of line joining origin of the ray to the centre of the sphere
	// Let us call this CL for centre line
	double CLx = centre.x - ray.origin.x ; 
	double CLy = centre.y - ray.origin.y ; 
	double CLz = centre.z - ray.origin.z ;

	// Compute the length of the component of CL along the ray CLR = CL . ray ( dot product )
	double CLR = ray.direction.dot( CLx, CLy, CLz ) ;
	
	if( CLR < 0 ) {
		//printf("\nDebug: Not heading\n") ;
		return false ; // Earliest possible rejection - ray not heading towards sphere
	}

	// If ( CLR - radius ) > ray.length we do an early rejection ( obviously )
	if( CLR - radius > ray.length ) {
		//printf("\nDebug: Too far case A ray.length = %lf\n", ray.length ) ;
		return false ;
	}

	// Otherwise go on to check if the ray actually intersects the sphere
	// Compute perpendicular distance of ray from centre of sphere
	double Csq = (CLx*CLx) + (CLy*CLy) + (CLz*CLz) ;
	double CRsq = Csq - (CLR*CLR) ;
	double inner_length_square = radius_sq - CRsq ;
	if( inner_length_square < 0 ) {
		//printf("\nDebug: Misses sphere %p\n", this ) ;
		return false ; // The ray does not intersect the sphere
	}	
	
	// Update length if this object is closer
	double new_length ; 
	if( Csq < radius_sq ) {
		// Inside sphere
		new_length = CLR + (double)sqrt( inner_length_square ) ;
		//cout << "\nDebug: Ray starting inside of sphere intersected sphere.\n" ;
		ray.inside = true ; r.inside = true ;
	}
	else { 
		//cout << "\nDebug: Ray starting outside of sphere intersected sphere.\n" ;
		new_length = CLR - (double)sqrt( inner_length_square ) ; // Outside the sphere
	}
	if( new_length < 0 || new_length > ray.length ) {
		//printf("\nDebug: Too far case B\n") ;
		ray.inside = r.inside = false ; return false ; // Failure - this object is further away
	}

	ray.length = new_length ;
	// Success - this object is closer
	ray.object = this ;
	r.object = this ;
	
	Vector3D point_of_intersection( ray.origin.x + ray.direction.x * ray.length,
					ray.origin.y + ray.direction.y * ray.length,
					ray.origin.z + ray.direction.z * ray.length 
				);
	if( !isTransformed ) r.length = ray.length ;
	else {
		Vector3D actual_point = transform_point( point_of_intersection ) ;
		r.length = ( Vector3D( actual_point.x - r.origin.x, 
				     actual_point.y - r.origin.y,
				     actual_point.z - r.origin.z 
				     ).length()
			   ) ;
	}
	
	// Compute u,v parametric coordinates for texture mapping (ONLY if necessary)
	if( surface -> texture )
	{
		/*
		 * Formulas for texture mapping a sphere 
		 * z = R * cos( pi * v )
		 * X = sqrt( R^2 - Z^2 ) * cos( 2 * pi * u )
		 * We need to be a little careful with signs, as always when it comes to inverse trigonometry :)
		 */
		double x = centre.x - point_of_intersection.x, y = centre.y - point_of_intersection.y, z = centre.z - point_of_intersection.z ;
		r.v = ( acos( z / radius ) ) / M_PI ;
		double discR = sqrt( radius_sq - z * z ) ;
		if( discR < x || discR < DELTA )
				r.u = 0.5 ; // Average
		else {
				r.u = ( acos( x / discR ) ) / ( 2 * M_PI ) ;
				if( y < 0 ) r.u = 1 - r.u ;
		}
	}
	
	//printf("\nDebug: Intersects sphere %p\n", this ) ;
	return true ;
}

// Object shader for a sphere
Color Sphere::shade( const Ray& ray, int depth )
{
	// Assume ray has already been intersected with the sphere - otherwise this will NOT be called
	Vector3D intersection(	ray.origin.x + ( ray.length * ray.direction.x ),
			      	ray.origin.y + ( ray.length * ray.direction.y ),
				ray.origin.z + ( ray.length * ray.direction.z )
			);
	
	Vector3D normal ;	
	if( isTransformed ) {
		// Transform intersection using inverse transform
		double result[3] ;
		for( int i = 0 ; i < 3 ; i++ )
		result[i] = inverse[i][0] * intersection.x +
			    inverse[i][1] * intersection.y +
		      	    inverse[i][2] * intersection.z +
			    inverse[i][3] ;
		
		Vector3D old_normal( result[0] - centre.x, result[1] - centre.y, result[2] - centre.z ) ;
		normal = transform_normal( old_normal ) ;
		// printf("\nDebug: Intersection = (%lf,%lf,%lf) Normal = (%lf,%lf,%lf) to sphere with centre (%lf,%lf,%lf)\n", intersection.x, intersection.y, intersection.z, normal.x, normal.y, normal.z, centre.x, centre.y, centre.z ) ;
	}
	else
	{
		normal = Vector3D( intersection.x - centre.x, intersection.y - centre.y, intersection.z - centre.z ) ;
	}

	normal.normalize() ;
	//printf("\nDebug: Normal at point (%f,%f,%f) to scaled sphere is (%f,%f,%f)\n", intersection.x, intersection.y, intersection.z, normal.x, normal.y, normal.z) ; 
	Vector3D view( -ray.direction.x, -ray.direction.y, -ray.direction.z ) ;
	
	// if texture mapped used texture shader
	if( surface -> texture ) {
		return surface -> texture_shade( ray.u, ray.v, intersection, normal, view, ray.n, ray.inside, depth ) ;
		
	}
	// Otherwise call surface shader with appropriate parameters
	//cout << "\nDebug: Calling surface shader from sphere " << this << " with centre ( " << centre.x << ", " << centre.y << ", " << centre.z << " )" << " and radius " << radius << "\n" ;
	else return surface -> shade( intersection, normal, view, ray.n, ray.inside, depth ) ;
}

// Returns an interval of intersection of the ray r with this sphere
vector <CSG_Interval*> Sphere::compute_intervals( const Ray& r )
{
	// Similar to the intersection routine to begin with
	Ray ray ;
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;

	// Compute direction ratios of line joining origin of the ray to the centre of the sphere
	// Let us call this CL for centre line
	double CLx = centre.x - ray.origin.x ; 
	double CLy = centre.y - ray.origin.y ; 
	double CLz = centre.z - ray.origin.z ;

	// Compute the length of the component of CL along the ray CLR = CL . ray ( dot product )
	double CLR = ray.direction.dot( CLx, CLy, CLz ) ;
	
	if( CLR < 0 ) {
		// Earliest possible rejection - ray not heading towards sphere
		vector <CSG_Interval*> new_list ;
		//printf("\nDebug: Ray does not intersect sphere with centre %f\n", centre.y ) ;
		return new_list ;
	}
	
	// Otherwise go on to check if the ray actually intersects the sphere
	// Compute perpendicular distance of ray from centre of sphere
	double Csq = (CLx*CLx) + (CLy*CLy) + (CLz*CLz) ;
	double CRsq = Csq - (CLR*CLR) ;
	double inner_length_square = radius_sq - CRsq ;
	if( inner_length_square < 0 ) { // The ray does not intersect the sphere
		vector <CSG_Interval*> new_list ;
		//printf("\nDebug: Ray does not intersect sphere with centre %f\n", centre.y ) ;
		return new_list ;
	}
	
	double tnear, tfar ;
	double inner_length = (double)sqrt( inner_length_square ) ;
	if( Csq < radius_sq ) { 
		// Inside sphere
		tfar = CLR + inner_length ;
		tnear = 0 ;
	}

	else {
		// Outside sphere
		tnear = CLR - inner_length ;
		tfar  = CLR + inner_length ;
	}

	if( isTransformed ) {
		// Transform back !
		if( !isZero( tnear ) ) {
			Vector3D nearpoint(ray.origin.x + tnear * ray.direction.x,
					ray.origin.y + tnear * ray.direction.y,	
					ray.origin.z + tnear * ray.direction.z
					);
			Vector3D transformed_point = transform_point( nearpoint ) ;
			tnear = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}

		if( !isZero( tfar ) ) {
			Vector3D farpoint(  ray.origin.x + tfar * ray.direction.x,
					ray.origin.y + tfar * ray.direction.y,	
					ray.origin.z + tfar * ray.direction.z
					);
			// Repeat calculation for far point of intersection
			Vector3D transformed_point = transform_point( farpoint ) ; 
			tfar = (Vector3D( transformed_point.x - r.origin.x,
						transformed_point.y - r.origin.y,	
						transformed_point.z - r.origin.z
						)).length() ;	
		}
	}
	CSG_Interval* new_interval = new CSG_Interval() ;
	new_interval -> tnear = tnear ; new_interval -> tfar = tfar ;
	new_interval -> pnear = new_interval -> pfar = this ;
	vector <CSG_Interval*> new_list ;
	new_list.push_back( new_interval ) ;
	//printf("\nDebug: Interval for sphere with centre %f = ( %f, %f )\n", centre.y, ( ( (new_list).size() == 0 ) ? 0 : (new_list)[0] -> tnear ), ( ( (new_list).size() == 0 ) ? 0 : (new_list)[0] -> tfar ) ) ;  
	return new_list ;
}

void Sphere::computeBounds( bVolume& b ) // Compute bounding volume for this sphere
{
	b.xmin = centre.x - radius ;
	b.xmax = centre.x + radius ;
	b.ymin = centre.y - radius ;
	b.ymax = centre.y + radius ;
	b.zmin = centre.z - radius ;
	b.zmax = centre.z + radius ;
	
	if( isTransformed ) {
		// Transform each of the vertices of the current bounding box using the same transformation
		// Take min x, min y, min z, max x, max y, max z
		transform_box( b ) ;
	}
}
