/*
 * Implements Whitted's adaptive supersampling algorithm
 * For antialiasing
 * Refer to chapter 15 Foley, Van Dam, Feiner and Hughes
 */

#include "color.h"
#include "ppm.h"
#include "surface.h"
#include "supersample.h"

#include <math.h>
#include <map>
using std::map ;

/* Some ray tracing methods imported from init.cpp 
 */
extern Color trace_pixel_color( double, double ) ;

/* Important notes
 * Use memoization to save on computation if possible
 */
map< Pixel, Color, compare_pixel > memoized_colors ;

/* Recursive adaptive subdivision routine
 * The center of the pixel is passed as a parameter
 * As is the size of the pixel
 * We also pass the values computed for the 4 corners of the pixel in an array
 * And a depth parameter to control explosion of the recursion
 */
Color adaptive_subdivide( Pixel p, Color corner_colors[4], int depth )
{
		// Decide whether to subdivide or not by using corner_colors
		Color mean( 0, 0, 0 ) ; int c ;
		for( c = 0 ; c < 4 ; c++ )
		{
				mean.red += corner_colors[c].red ;
				mean.green += corner_colors[c].green ;
				mean.blue += corner_colors[c].blue ;
				
		}
		mean.red /= 4 ; mean.green /= 4 ; mean.blue /= 4 ;
		
		if( depth > MAX_AA_DEPTH ) return mean ; 
		bool subdivide_flag = false ;
		for( c = 0 ; c < 4 ; c++ )
		{
			if( fabs( corner_colors[c].red - mean.red ) > AA_QUALITY )	{
				//printf("\nDebug: At size %lf Corner %d red = %lf Mean red = %lf\n", p.size, c, corner_colors[c].red, mean.red ) ;
				subdivide_flag = true ; break ;
			}
			
			if( fabs( corner_colors[c].green - mean.green ) > AA_QUALITY )	{
				//printf("\nDebug: At size %lf Corner %d green = %lf Mean green = %lf\n", p.size, c, corner_colors[c].green, mean.green ) ;
				subdivide_flag = true ; break ;
			}
			
			if( fabs( corner_colors[c].blue - mean.blue ) > AA_QUALITY )	{
				//printf("\nDebug: At size %lf Corner %d blue = %lf Mean blue = %lf\n", p.size, c, corner_colors[c].blue, mean.blue ) ;
				subdivide_flag = true ; break ;
			}
		}
		
		// If we do not need to subdivide simply return mean
		if( !subdivide_flag ) 
		{
				return mean ;
		}
		
		// Shoot rays through the midpoints of sides of the pixel
		Pixel left_midpoint, top_midpoint, right_midpoint, bottom_midpoint ;
		left_midpoint.size = top_midpoint.size = right_midpoint.size = bottom_midpoint.size = 0.5 * p.size ;
		left_midpoint.i = p.i - 0.5 * p.size ; 
		right_midpoint.i = p.i + 0.5 * p.size ;
		top_midpoint.j = p.j - 0.5 * p.size ;
		bottom_midpoint.j = p.j + 0.5 * p.size ;
		left_midpoint.j = right_midpoint.j = p.j ;
		top_midpoint.i = bottom_midpoint.i = p.i ;
		
		Color left, top, right, bottom, centre ;
		
		centre = trace_pixel_color( p.i, p.j ) ;
		left = trace_pixel_color( left_midpoint.i, left_midpoint.j ) ;
		top = trace_pixel_color( top_midpoint.i, top_midpoint.j ) ;
		right = trace_pixel_color( right_midpoint.i, right_midpoint.j ) ;
		bottom = trace_pixel_color( bottom_midpoint.i, bottom_midpoint.j ) ;
		
		// Now that we have all the colors, do recursive adaptive subdivision where necessary
		
		Color quadrants[4], corners[4] ; Pixel centre_pixel ;
		
		// Top left quadrant
		corners[0] = corner_colors[0] ; corners[1] = top ; corners[2] = centre ; corners[3] = left ;
		centre_pixel.i = p.i - 0.25 * p.size ; centre_pixel.j = p.j - 0.25 * p.size ; centre_pixel.size = 0.5 * p.size ;
		quadrants[0] = adaptive_subdivide( centre_pixel, corners, 1 + depth ) ;
		
		// Top right quadrant
		corners[0] = top ; corners[1] = corner_colors[1] ; corners[2] = right ; corners[3] = centre ;
		centre_pixel.i = p.i + 0.25 * p.size ; centre_pixel.j = p.j - 0.25 * p.size ; centre_pixel.size = 0.5 * p.size ;
		quadrants[1] = adaptive_subdivide( centre_pixel, corners, 1 + depth ) ;
		
		// Bottom left quadrant
		corners[0] = left ; corners[1] = centre ; corners[2] = bottom ; corners[3] = corner_colors[3] ;
		centre_pixel.i = p.i - 0.25 * p.size ; centre_pixel.j = p.j + 0.25 * p.size ; centre_pixel.size = 0.5 * p.size ;
		quadrants[2] = adaptive_subdivide( centre_pixel, corners, 1 + depth ) ;
		
		// Bottom right quadrant
		corners[0] = centre ; corners[1] = right ; corners[2] = corner_colors[2] ; corners[3] = bottom ;
		centre_pixel.i = p.i + 0.25 * p.size ; centre_pixel.j = p.j + 0.25 * p.size ; centre_pixel.size = 0.5 * p.size ;
		quadrants[3] = adaptive_subdivide( centre_pixel, corners, 1 + depth ) ;

		// Average quadrant colors
		Color result( 0, 0, 0 ) ;
		for( c = 0 ; c < 4 ; c++ ) {
				result.red += quadrants[c].red ;
				result.green += quadrants[c].green ;
				result.blue += quadrants[c].blue ;
		}
		result.red /= 4 ; result.green /= 4 ; result.blue /= 4 ;
		return result ;
}
