// class definition for class Surface - surface.cpp
// A good portion of the ray tracer is implemented here

#include "surface.h"
#include "primitive.h"
#include "texture.h"
#include "arealight.h"
#include <assert.h>
#include <iostream>
#include <math.h>
#include <stdlib.h> // For random numbers

using std::cout ;
using std::vector ;

extern int global_texture_filter_setting ;
extern Color background ;
extern vector<Light*> lightList ;
extern vector<AreaLight*> areaLightList ;
extern bool enable_path ;

/*
 * Definitions for a generic surface
 */
Surface::Surface()
{
	texture = NULL ;
}

Surface::Surface( const Surface& s )
{
	texture = s.texture ;
}

Color Surface::shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth ) 
{
	// Should not ever be called - to satisfy g++
	return Color( 0, 0, 0 ) ;
}	

Color Surface::texture_shade( const double u, const double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth ) 
{
	// Should not ever be called - to satisfy g++
	return Color( 0, 0, 0 ) ;
}	

/*
 * Definitions for a surface using Phong illumination
 */
PhongSurface::PhongSurface( Color col, double KA, double KD, double KS, double NS, double KR, double KT, double MU ) : Surface()
{
	intrinsic_color = col ;
	ka = KA ;
	kd = KD ;
	ks = KS ;
	ns = NS ;
	kr = KR ;
	kt = KT ;
	n = MU ;
}

PhongSurface::PhongSurface( const PhongSurface& s ) : Surface()
{
	intrinsic_color = s.intrinsic_color;
	ka = s.ka ;
	kd = s.kd ;
	ks = s.ks ;
	ns = s.ns ;
	kr = s.kr ;
	kt = s.kt ;
	n = s.n ;
	texture = s.texture ;
}

void Surface::setTexture( Texture* t )
{
	texture = t ;
}

// Compute negative of original incident ray given refracted ray and a few parameters with correct signs
// TODO - check and fix signs, if required
Vector3D Surface::refract( double cos_r, double sin_i, double cos_i, const Vector3D& normal, const Vector3D& view )
{
	Vector3D transverse = view.cross( normal ) ;
	transverse = normal.cross( transverse ) ;
	transverse.normalize() ;
	Vector3D result ;
	result.x = -normal.x * cos_i - sin_i * transverse.x ;
	result.y = -normal.y * cos_i - sin_i * transverse.y ;
	result.z = -normal.z * cos_i - sin_i * transverse.z ;
	result.normalize() ;
	return result ;
}

// Shades a texture mapped surface
// A separate method is used to save on stack space for non texture mapped objects
Color PhongSurface::texture_shade( const double u, const double v,const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth )
{
	// The surface is texture mapped, use the texture map to recompute the intrinsic color in each of these calls
	intrinsic_color = texture -> map( u, v, global_texture_filter_setting ) ;
	return shade( intersection, normal, view, original_n, inside, depth ) ;
}

// Shades a phong surface given intersection pt, normal, view vector, lights, other objects and refractive index of original medium
// Handles three kinds of light sources currently - ambient, point and directional
// Uses Phong model for specular reflection (currently)
Color PhongSurface::shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth )
{	
	bool specular_flag = true ;
	// Compute contribution due to each of the light sources to each color: r,g,b
	double total_red = 0, total_green = 0, total_blue = 0 ;
	
	if( !enable_path ) {
		// Ordinary ray tracing and phong model
		for( vector<Light*>::const_iterator it = lightList.begin() ; it != lightList.end() ; it++ ) {
			Light* lptr = *it ;
			// We have ambient light sources to start with
			if( lptr -> getLightType() == AMBIENT ) {
				total_red += ka * intrinsic_color.red * ( lptr -> red ) ;
				total_green += ka * intrinsic_color.green * ( lptr -> green ) ;
				total_blue += ka * intrinsic_color.blue * ( lptr -> blue ) ;
			}

			else {
				// Get the direction of the unit vector from the light source to the point of intersection 
				// In the case of a point source, this is simply the line joining them
				// For a directional source, this is the direction of the light

				Vector3D incidence ; 

				// The max distance upto which we trace shadow rays - see subtle point below
				double shadow_length = MAX_LENGTH ;

				if( lptr -> getLightType() == DIRECTIONAL )
					incidence = Vector3D::normalize( lptr -> get_light_vector() ) ;
				else /* if( lptr -> lighttype == POINT ) */ {
					incidence.x = intersection.x - ( lptr -> get_light_vector() ).x ;
					incidence.y = intersection.y - ( lptr -> get_light_vector() ).y ;
					incidence.z = intersection.z - ( lptr -> get_light_vector() ).z ;
					shadow_length = incidence.length() ;
					incidence.normalize() ;
				}

				// Before rushing into calculating anything stupidly,
				// First determine if the intersection point in a shadow
				// The simplest method to do this is a little hack :)
				// Move away from the intersection point a distance +delta
				// along the incidence vector ( i.e towards the light source )
				// From this point, draw a new ray towards the light source and determine if 
				// it intersects any objects "before" it reaches the light source.
				// The above point is rather important and tends to be missed
				Vector3D reverse( -incidence.x, -incidence.y, -incidence.z ) ;
				Vector3D point_above_surface ;
				point_above_surface.x = intersection.x - ( incidence.x * DELTA ) ;
				point_above_surface.y = intersection.y - ( incidence.y * DELTA ) ;
				point_above_surface.z = intersection.z - ( incidence.z * DELTA ) ;

				// Look from a point slightly "above" the surface towards light source
				// Be careful with double !
				Ray shadow_ray( point_above_surface, reverse ) ;	

				shadow_ray.length = shadow_length ; // Objects behind light source excluded

				// Use hack - trace shadow ray to see if there is an object blocking the light
				// source 	
				if( shadow_ray.shadow_trace() )
				{
					// No illumination due to this light source - continue
					//Primitive* pptr = dynamic_cast <Primitive*> (shadow_ray.object) ;
					//if( !pptr ) {
					// CSG Object
					//	printf("\nDebug: Shadow due to CSG Object with shadow ray from (%lf,%lf,%lf) above intersection (%lf,%lf,%lf) in direction (%lf,%lf,%lf)", shadow_ray.origin.x, shadow_ray.origin.y, shadow_ray.origin.z, intersection.x, intersection.y, intersection.z, shadow_ray.direction.x, shadow_ray.direction.y, shadow_ray.direction.z ) ;
					//	}
					//	else {
					//Color surface_color = 
					//( ( ( Primitive* ) ( shadow_ray.object ) ) -> surface ).intrinsic_color ;				
					// if( fabs(intrinsic_color.red - 0.822) < DELTA )
					//printf("\nDebug: Shadow due to (%lf,%lf,%lf) with shadow ray from (%lf,%lf,%lf) in direction (%lf,%lf,%lf) having length limit %lf but object at distance %lf\n", surface_color.red, surface_color.green, surface_color.blue, shadow_ray.origin.x, shadow_ray.origin.y, shadow_ray.origin.z, shadow_ray.direction.x, shadow_ray.direction.y, shadow_ray.direction.z, shadow_length, shadow_ray.length ) ;
					//	}
					continue ;
				}

				//printf("\nDebug: No shadow\n") ;

				// Now that we have got this unit vector, determine the contribution from 
				// a) diffuse b) specular reflection

				// First compute component of incidence unit vector along the normal cos theta
				// And use lambert's illumination
				double lambert_factor = reverse.dot( normal ) ;

				if( lambert_factor > 0 ) {
					// Diffuse reflection
					total_red += kd * intrinsic_color.red * lambert_factor * ( lptr -> red ) ; 
					total_green += kd * intrinsic_color.green * lambert_factor * ( lptr -> green ) ; 
					total_blue += kd * intrinsic_color.blue * lambert_factor * ( lptr -> blue ) ;
				}

				// Specular reflection - use Phong model for now
				// Compute the angle made by the ray to the eye with the actual reflected ray
				Vector3D reflected_ray( incidence.x + ( 2 * lambert_factor * normal.x ),
						incidence.y + ( 2 * lambert_factor * normal.y ),
						incidence.z + ( 2 * lambert_factor * normal.z )
						);

				double azimuthal_cosine = view.dot( reflected_ray ) ; 	

				if( azimuthal_cosine > 0 ) {
					// Now add Ks * ( ( cos a ) ^ shininess )
					double specular_light = ks * pow( azimuthal_cosine, ns ) ;
					total_red += specular_light * ( lptr -> red ) ; 
					total_green += specular_light * ( lptr -> green ) ; 
					total_blue += specular_light * ( lptr -> blue ) ;
				}
			}	// end of else
		}	// end of loop over light sources
	}

	else {
		specular_flag = false ;
		// Path tracing
		// First loop over area lights
		for( int i = 0 ; i < areaLightList.size() ; i++ )
		{
			AreaLight* AL = areaLightList[i] ;
			// Test visibility of random position on area source
			double chi1 = rand_double() ; double chi2 = rand_double() ;
			Vector3D random_position ;
			switch( AL -> orientation ) {
				case POSITIVE_X:
				case NEGATIVE_X:	
					random_position.x = AL -> main ;
				       	random_position.y = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
				       	random_position.z = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
				case POSITIVE_Y:
				case NEGATIVE_Y:	
					random_position.y = AL -> main ;
				       	random_position.z = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
				       	random_position.x = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
				case POSITIVE_Z:
				case NEGATIVE_Z:	
					random_position.z = AL -> main ;
				       	random_position.x = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
				       	random_position.y = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
			}

			Vector3D direction( random_position.x - intersection.x, random_position.y - intersection.y, random_position.z - intersection.z ) ;
			direction.normalize() ;
			Vector3D point_above_surface( intersection.x + DELTA * direction.x, intersection.y + DELTA * direction.y, intersection.z + DELTA * direction.z ) ;
			Ray random_ray( point_above_surface, direction ) ;
			random_ray.trace() ;
			if( random_ray.object == AL )
			{
				// Add direct illumination
				// Only diffuse reflectance is considered obviously
				double costheta = fabs( direction.dot( normal ) ) ;
				total_red += kd * costheta * intrinsic_color.red * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> red ) ;
				total_green += kd * costheta * intrinsic_color.green * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> green ) ;
				total_blue += kd * costheta * intrinsic_color.blue * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> blue ) ;
				
				/*
				// Also add the phong spot
				// Compute the angle made by the ray to the eye with the actual reflected ray
				Vector3D reflected_ray( -direction.x + ( 2 * costheta * normal.x ),
						-direction.y + ( 2 * costheta * normal.y ),
						-direction.z + ( 2 * costheta * normal.z )
						);

				double azimuthal_cosine = view.dot( reflected_ray ) ; 	

				if( azimuthal_cosine > 0 ) {
					// Now add Ks * ( ( cos a ) ^ shininess )
					double specular_light = ks * pow( azimuthal_cosine, ns ) ;
					total_red += specular_light * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> red ) ; 
					total_green += specular_light *( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> green ) ;  
					total_blue += specular_light * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> blue ) ;
				//	printf("\nDebug: red = %lf green = %lf blue = %lf\n", total_red, total_green, total_blue ) ;
				}
				*/
			}
		}	// End of loop over area light sources

		// Toss a coin for diffuse, specular
		double chi = rand_double() ;
		if( chi < kd )
		{
			// Trace a diffuse ray
			// Trace diffuse ray by importance sampling
			// Choose diffuse ray direction
			double chi1 = rand_double() ; double chi2 = rand_double() ;
			double theta = acos( sqrt( chi1 ) ) ;		
			double phi = 2 * M_PI * chi2 ;
			double costheta, sintheta, cosphi, sinphi ;
			costheta = cos( theta ) ; sintheta = sin( theta ) ;
			cosphi = cos( phi ) ; sinphi = sin( phi ) ;
			
			Vector3D plane_vec1, plane_vec2 ;
			// Choose a direction perpendicular to n
			if( normal.y < DELTA && normal.z < DELTA ) {
				plane_vec1 = Vector3D( 0, 1, 0 ) ;
				plane_vec2 = Vector3D( 0, 0, 1 ) ;
			}
			else {
				plane_vec1 = Vector3D( 0, normal.z, -normal.y ) ;
				plane_vec1.normalize() ;
				plane_vec2 = normal ;
				plane_vec2.cross( plane_vec1 ) ;
				plane_vec2.normalize() ;
			}
			Vector3D diffuse_direction( 
					costheta * normal.x + sintheta * cosphi * plane_vec1.x + sintheta * sinphi * plane_vec2.x,
					costheta * normal.y + sintheta * cosphi * plane_vec1.y + sintheta * sinphi * plane_vec2.y,	
					costheta * normal.z + sintheta * cosphi * plane_vec1.z + sintheta * sinphi * plane_vec2.z 
					) ;
			Vector3D start_of_diffuse_ray( intersection.x + DELTA * diffuse_direction.x,
					intersection.y + DELTA * diffuse_direction.y,
					intersection.z + DELTA * diffuse_direction.z
					);
			Ray diffuse_ray( start_of_diffuse_ray, diffuse_direction ) ;	
			diffuse_ray.n = original_n ; // Starts right here
			
			if( diffuse_ray.trace() )
			{
				Color diffuse_color = diffuse_ray.shade( depth + 1 ) ;
				total_red += kd * intrinsic_color.red * diffuse_color.red * costheta ;
				total_green += kd * intrinsic_color.green * diffuse_color.green * costheta ;
				total_blue += kd * intrinsic_color.blue * diffuse_color.blue * costheta ;
				//printf("\nDebug: Adding diffuse contribution %lf,%lf,%lf at depth %d from object %p\n", diffuse_color.red, diffuse_color.green, diffuse_color.blue, depth + 1, diffuse_ray.object ) ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}

			else {
				//printf("\nDebug: Adding diffuse contribution of background at depth %d\n", depth + 1) ;
				total_red += kd * intrinsic_color.red * background.red * costheta ;
				total_green += kd * intrinsic_color.green * background.green * costheta ;
				total_blue += kd * intrinsic_color.blue * background.blue * costheta ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}	
		}

		else {
			// Specular
			specular_flag = true ;
		}
	}
	
	if( specular_flag ) {
		// Contribution of light sources has been considered
		// Now look at contribution from reflected and refracted rays

		// Reflection first
		// A ray from an object reflected off the surface onto the eye is what we seek
		// So we take the ray from the eye to the surface as the incident ray
		// And compute the reflected ray, and now see if it intersects any object
		Color reflected_color ;	

		// Some refraction related stuff first for correctness :)
		// If there is total internal reflection, add the rest ( 1 - Kr ) from the reflected ray's contribution
		// Check for t.i.r
		double cos_view_angle = view.dot( normal ) ;
		double sin_r = sqrt( 1 - ( cos_view_angle * cos_view_angle ) ) ;

		// We only consider the special case when there are only two media possible: air and the object.
		// More complicated cases: TODO
		// Get n1, n2 - find which is which - for this use the argument original_n

		double n1, n2 = original_n ; // The refracted ray in n2 is what meets the eye :)
		if( fabs( n2 - 1 ) < DELTA ) 
			n1 = n ;
		else n1 = 1 ;
		
		if( ( kr > 0 || sin_r > ( n1 / n2 ) ) ) { 
			Vector3D reflect_dir(	-view.x + ( 2 * cos_view_angle * normal.x ), 
					-view.y + ( 2 * cos_view_angle * normal.y ),
					-view.z + ( 2 * cos_view_angle * normal.z )
					) ;

			// Start a new ray a little ( delta ) above the point of intersection along this direction
			Vector3D start_of_reflected_ray( intersection.x + DELTA*reflect_dir.x,
					intersection.y + DELTA*reflect_dir.y,
					intersection.z + DELTA*reflect_dir.z
					) ;

			Ray reflected_ray( start_of_reflected_ray, reflect_dir ) ;
			reflected_ray.n = original_n ; // Starts right here
			// Ray trace all the objects recursively using the reflected ray and get the color 
			// which is ( presumably ) reflected / refracted off another object's surface
			// On the other hand, remember that if no object intersects the reflected ray,
			// We still need to add in a contribution equal to the background color
			//printf("\nDebug: Spawning reflected ray with origin (%lf,%lf,%lf) in direction (%lf,%lf,%lf)\n", reflected_ray.origin.x, reflected_ray.origin.y, reflected_ray.origin.z, reflected_ray.direction.x, reflected_ray.direction.y, reflected_ray.direction.z) ;
			if( reflected_ray.trace() )
			{
				// Now add a contribution from the color returned by the shading routine
				// Aha! - recursion
				reflected_color = reflected_ray.shade( depth+1 ) ;
				// Add in a contribution scaled down in intensity by the reflectance Kr
				total_red += kr * reflected_color.red ;
				total_green += kr * reflected_color.green ;
				total_blue += kr * reflected_color.blue ;
			}

			else
			{
				// Add in a background contribution scaled down by Kr
				total_red += kr * background.red ;
				total_green += kr * background.green ;
				total_blue += kr * background.blue ;
			}

			if( sin_r > ( n1 / n2 ) ) {
				// Total internal reflection
				// Add a ( 1 - Kr ) * reflection color that was already computed	
				//cout << "\nDebug: Total internal reflection taking place at depth " << depth + 1 << " contributing color = " << reflected_color.red << " " << reflected_color.green << " " << reflected_color.blue << "\n" ;
				total_red += reflected_color.red * ( 1 - kr ) ;	
				total_green += reflected_color.green * ( 1 - kr ) ;	
				total_blue += reflected_color.blue * ( 1 - kr ) ;	
			}
		}
		

		// Now refraction needs to be handled
		// This is not really very different from reflection
		// We need to extend the ray class to include an additional attribute
		// This is the refractive index of the medium the ray is travelling in
		// We have already considered total internal reflection
		if( kt > 0 && ( sin_r < ( n1 / n2 ) ) )
		{
			// Original incident ray needs to be computed
			double sin_i = ( sin_r * n2 / n1 ) ;
			double cos_i = sqrt( 1 - ( sin_i * sin_i ) ) ;
			if( cos_view_angle < 0 ) cos_i = -cos_i ;

			Vector3D original_ray = refract( cos_view_angle, sin_i, cos_i, normal, Vector3D::normalize(view) ) ;
			Vector3D point_below_surface( 	intersection.x + DELTA*original_ray.x,		       
					intersection.y + DELTA*original_ray.y,		       
					intersection.z + DELTA*original_ray.z	
					) ;

			// Draw ray
			Ray refracted_ray( point_below_surface, original_ray ) ;
			refracted_ray.n = n1 ; // Set it to the 'other' refractive index 
			// Trace it recursively

			if( refracted_ray.trace() )
			{
				Color refracted_color = refracted_ray.shade( depth+1 ) ;
				//printf("\nDebug: Refracted Color = %lf,%lf,%lf at depth %d from object %p\n", refracted_color.red, refracted_color.green, refracted_color.blue, depth + 1, refracted_ray.object ) ; 
				total_red += kt * refracted_color.red ;
				total_green += kt * refracted_color.green ;
				total_blue += kt * refracted_color.blue ;
			}

			else // It's transparent, stupid!
			{
				//printf("\nDebug: Refracted Color = background at depth %d\n", depth + 1 ) ; 
				total_red += kt * background.red ;
				total_green += kt * background.green ;
				total_blue += kt * background.blue ;
			}	
		}
	}
	
	// We have factored in plenty of contributions to the total color
	if( total_red > 1.0 ) total_red = 1 ;	
	if( total_green > 1.0 ) total_green = 1 ;	
	if( total_blue > 1.0 ) total_blue = 1 ;	

	//printf("\nDebug: Color of surface = %lf,%lf,%lf\n", total_red, total_green, total_blue ) ;
	return Color( total_red, total_green, total_blue );
}

EmissiveSurface::EmissiveSurface( Color c )
{
	red = c.red ;
	green = c.green ;
	blue = c.blue ;
}

EmissiveSurface::EmissiveSurface( const EmissiveSurface& e )
{
	red = e.red ;
	green = e.green ;
	blue = e.blue ;
}

// Very simplified model of an emissive surface follows - embarassingly simple
Color EmissiveSurface::shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth ) 
{
	Color result( 0, 0, 0 ) ;
	double lambert = view.dot( normal ) ;
	if( lambert > 0 ) {
		result.red = red ;	
		result.green = green ;	
		result.blue = blue ;	
	}
	return result ;
}

Color EmissiveSurface::texture_shade( double u, double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth ) 
{
	// Textures - nice try
	return shade( intersection, normal, view, original_n, inside, depth ) ;
}

SchlickSurface::SchlickSurface( double F, double sig, double r, double t, double mu)
{
	Fo = F ;
	sigma = sig ;
	kr = r ;
	kt = t ;
	n = mu ;
	g = 4 * sigma * ( 1 - sigma ) ;

	if( sigma < 0.5 ) {
		d = 0 ; s = 1 - g ;
	}

	else {
		d = 1 - g ; s = 0 ; 
	}
}

SchlickSurface::SchlickSurface( const SchlickSurface& ss )
{
	Fo = ss.Fo ;
	sigma = ss.sigma ;
	kr = ss.kr ;
	kt = ss.kt ;
	n = ss.n ;
	g = 4 * s * ( 1 - s ) ;

	if( sigma < 0.5 ) {
		d = 0 ; s = 1 - g ;
	}

	else {
		d = 1 - g ; s = 0 ; 
	}
}

// Generates a random double between 0 and 1
double rand_double()
{
	return (  (double) abs( rand() ) )  / RAND_MAX ; 
}

// The hard part - Monte Carlo shade algorithm for a Schlick BRDF
Color SchlickSurface::shade( const Vector3D& intersection, const Vector3D& norm, const Vector3D& view, const double original_n, bool inside, int depth )
{
	Vector3D normal = norm ;
	double v = view.dot( normal ) ; 
	if( v < 0 ) {
		normal.x = -normal.x ;
		normal.y = -normal.y ;
		normal.z = -normal.z ;
		v = -v ;
	}
	double total_red = 0, total_green = 0, total_blue = 0 ;

	double chi, chi1, chi2 ; // Russian roulette numbers

	Vector3D H ; // The halfway vector
	// The specular factor S(u) 
	double u, Su ;
	
	if( d > 0 ) {
		// First loop over area lights
		for( int i = 0 ; i < areaLightList.size() ; i++ )
		{
			AreaLight* AL = areaLightList[i] ;
			// Test visibility of random position on area source
			double chi1 = rand_double() ; double chi2 = rand_double() ;
			Vector3D random_position ;
			switch( AL -> orientation ) {
				case POSITIVE_X:
				case NEGATIVE_X:	
					random_position.x = AL -> main ;
					random_position.y = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
					random_position.z = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
				case POSITIVE_Y:
				case NEGATIVE_Y:	
					random_position.y = AL -> main ;
					random_position.z = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
					random_position.x = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
				case POSITIVE_Z:
				case NEGATIVE_Z:	
					random_position.z = AL -> main ;
					random_position.x = ( AL -> min1 ) * chi1 + ( AL -> max1 ) * ( 1 - chi1 ) ;	
					random_position.y = ( AL -> min2 ) * chi2 + ( AL -> max2 ) * ( 1 - chi2 ) ;
					break ;	
			}

			Vector3D direction( random_position.x - intersection.x, random_position.y - intersection.y, random_position.z - intersection.z ) ;
			direction.normalize() ;
			Vector3D point_above_surface( intersection.x + DELTA * direction.x, intersection.y + DELTA * direction.y, intersection.z + DELTA * direction.z ) ;
			Ray random_ray( point_above_surface, direction ) ;
			random_ray.trace() ;

			H.x = direction.x + view.x ;	
			H.y = direction.y + view.y ;	
			H.z = direction.z + view.z ;	
			H.normalize() ;
			u = view.dot( H ) ; Su = Fo + ( 1 - Fo ) * ( pow( 1 - u, 5 ) ) ;
			if( random_ray.object == AL )
			{
				// Add direct illumination
				// Only diffuse reflectance is considered obviously
				//printf("\nDebug: Added direct illumination at depth %d\n", depth );
				double costheta = fabs( direction.dot( normal ) ) ;
				total_red += Su / M_PI * costheta * d * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> red ) ;
				total_green += Su / M_PI * costheta * d * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> green ) ;
				total_blue += Su / M_PI * costheta * d * ( ( dynamic_cast<EmissiveSurface*>( AL -> surface ) ) -> blue ) ;
			}
		}	// End of loop over area light sources
	}
	// Only emissive surfaces are supported in path tracing, obviously :)
	// First choose one: diffuse, specular or glossy
	// Using Russian roulette
	chi = rand_double() ; 
	if( chi < d ) {

		// Trace diffuse ray by importance sampling
		// Choose diffuse ray direction
		chi1 = rand_double() ; chi2 = rand_double() ;
		double theta = acos( sqrt( chi1 ) ) ;		
		double phi = 2 * M_PI * chi2 ;
		double costheta, sintheta, cosphi, sinphi ;
		costheta = cos( theta ) ; sintheta = sin( theta ) ;
		cosphi = cos( phi ) ; sinphi = sin( phi ) ;

		// Choose a direction perpendicular to n
		Vector3D plane_vec1( 0, normal.z, -normal.y ) ;
		Vector3D plane_vec2( normal.x * normal.x - 1, normal.x * normal.y, normal.x * normal.z ) ;
		Vector3D diffuse_direction( 
				costheta * normal.x + sintheta * sinphi * plane_vec2.x,
				costheta * normal.y + sintheta * cosphi * plane_vec1.y + sintheta * sinphi * plane_vec2.y,	
				costheta * normal.z + sintheta * cosphi * plane_vec1.z + sintheta * sinphi * plane_vec2.z 
				) ;
		Vector3D start_of_diffuse_ray( intersection.x + DELTA * diffuse_direction.x,
				intersection.y + DELTA * diffuse_direction.y,
				intersection.z + DELTA * diffuse_direction.z
				);
		Ray diffuse_ray( start_of_diffuse_ray, diffuse_direction ) ;	
		diffuse_ray.n = original_n ; // Starts right here

		H.x = diffuse_direction.x + view.x ;	
		H.y = diffuse_direction.y + view.y ;	
		H.z = diffuse_direction.z + view.z ;	
		H.normalize() ;
		u = view.dot( H ) ; Su = Fo + ( 1 - Fo ) * ( pow( 1 - u, 5 ) ) ;
		if( diffuse_ray.trace() )
		{
			Color diffuse_color = diffuse_ray.shade( depth + 1 ) ;
			total_red += Su / M_PI * diffuse_color.red * costheta ;
			total_green += Su / M_PI * diffuse_color.green * costheta ; 	
			total_blue += Su / M_PI * diffuse_color.blue * costheta ; 	
			if( total_red > 1.0 ) total_red = 1.0 ;
			if( total_green > 1.0 ) total_green = 1.0 ;
			if( total_blue > 1.0 ) total_blue = 1.0 ;
			return Color( total_red, total_green, total_blue ) ;
		}
		else {
			total_red += Su / M_PI * background.red * costheta ;
			total_green += Su / M_PI * background.green * costheta ; 	
			total_blue += Su / M_PI * background.blue * costheta ; 	
			if( total_red > 1.0 ) total_red = 1.0 ;
			if( total_green > 1.0 ) total_green = 1.0 ;
			if( total_blue > 1.0 ) total_blue = 1.0 ;
			return Color( total_red, total_green, total_blue ) ;
		}	
	}

	else if( chi < s + d ) {
		// Trace specular reflected or refracted ray
		double cos_view_angle = view.dot( normal ) ;
		double sin_r = sqrt( 1 - ( cos_view_angle * cos_view_angle ) ) ;
		double n1, n2 = original_n ; 
		if( fabs( n2 - 1 ) < DELTA ) 
			n1 = n ;
		else n1 = 1 ;
		if( ( kr > 0 || sin_r > ( n1 / n2 ) ) ) { 
			Vector3D reflect_dir(	-view.x + ( 2 * cos_view_angle * normal.x ), 
					-view.y + ( 2 * cos_view_angle * normal.y ),
					-view.z + ( 2 * cos_view_angle * normal.z )
					) ;

			Vector3D start_of_reflected_ray( intersection.x + DELTA*reflect_dir.x,
					intersection.y + DELTA*reflect_dir.y,
					intersection.z + DELTA*reflect_dir.z
					) ;

			Ray reflected_ray( start_of_reflected_ray, reflect_dir ) ;
			reflected_ray.n = original_n ; // Starts right here
			double costheta = v ; // Do NOT waste effort  Remember the law!
			H.x = reflect_dir.x + view.x ;	
			H.y = reflect_dir.y + view.y ;	
			H.z = reflect_dir.z + view.z ;	
			H.normalize() ;
			u = view.dot( H ) ; Su = Fo + ( 1 - Fo ) * ( pow( 1 - u, 5 ) ) ;
			Color reflected_color( 0, 0, 0 ) ;
			if( reflected_ray.trace() )
			{
				reflected_color = reflected_ray.shade( depth+1 ) ;
				total_red += Su * 2 * kr * reflected_color.red ;
				total_green += Su * 2 * kr * reflected_color.green ;
				total_blue += Su * 2 * kr * reflected_color.blue ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}

			else
			{
				total_red += 2 * Su * kr * background.red ;
				total_green += 2 * Su * kr * background.green ;
				total_blue += 2 * Su * kr * background.blue ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}

			if( sin_r > ( n1 / n2 ) ) {
				// Total internal reflection
				total_red += 2 * Su * reflected_color.red * ( 1 - kr ) ;	
				total_green += 2 * Su * reflected_color.green * ( 1 - kr ) ;	
				total_blue += 2 * Su * reflected_color.blue * ( 1 - kr ) ;	
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}
		}

		// Refraction
		else if( kt > 0 && ( sin_r < ( n1 / n2 ) ) ) 
		{
			double sin_i = ( sin_r * n2 / n1 ) ;
			double cos_i = sqrt( 1 - ( sin_i * sin_i ) ) ;
			double costheta = cos_i ;
			if( cos_view_angle < 0 ) cos_i = -cos_i ;

			Vector3D original_ray = refract( cos_view_angle, sin_i, cos_i, normal, Vector3D::normalize(view) ) ;
			Vector3D point_below_surface( 	intersection.x + DELTA*original_ray.x,		       
					intersection.y + DELTA*original_ray.y,		       
					intersection.z + DELTA*original_ray.z	
					) ;

			// Draw ray
			Ray refracted_ray( point_below_surface, original_ray ) ;
			refracted_ray.n = n1 ; // Set it to the 'other' refractive index 
			// Trace it recursively

			H.x = original_ray.x + view.x ;	
			H.y = original_ray.y + view.y ;	
			H.z = original_ray.z + view.z ;	
			H.normalize() ;
			u = view.dot( H ) ; Su = Fo + ( 1 - Fo ) * ( pow( 1 - u, 5 ) ) ;
			if( refracted_ray.trace() )
			{
				Color refracted_color = refracted_ray.shade( depth+1 ) ;
				//printf("\nDebug: Adding refracted contribution %lf,%lf,%lf at depth %d from object %p\n", refracted_color.red, refracted_color.green, refracted_color.blue, depth + 1, refracted_ray.object ) ;
				total_red += 2 * Su * kt * refracted_color.red ;
				total_green += 2 * Su * kt * refracted_color.green ;
				total_blue += 2 * Su * kt * refracted_color.blue ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}

			else // It's transparent, stupid!
			{
				total_red += 2 * kt * background.red ;
				total_green += 2 * kt * background.green ;
				total_blue += 2 * kt * background.blue ;
				if( total_red > 1.0 ) total_red = 1.0 ;
				if( total_green > 1.0 ) total_green = 1.0 ;
				if( total_blue > 1.0 ) total_blue = 1.0 ;
				return Color( total_red, total_green, total_blue ) ;
			}	
		}
	}	

	else {
		// Trace glossy ray
		double chi1 = rand_double() ;
		double chi2 = rand_double() ;
		double t = ( sqrt( chi1 ) - chi1 ) * sigma + chi1 ;
		double ts = sqrt( 1 - t * t ) ;
		double cosphi = cos( chi2 * 2 * M_PI ) ; // Isotropic
		double sinphi = sin( chi2 * 2 * M_PI ) ; // Isotropic
		Vector3D plane_vec1( 0, normal.z, -normal.y ) ;
		Vector3D plane_vec2( normal.x * normal.x - 1, normal.x * normal.y, normal.x * normal.z ) ;
		H.x = t * normal.x + ts * sinphi * plane_vec2.x  ; 
		H.y = t * normal.y + ts * cosphi * plane_vec1.y + ts * sinphi * plane_vec2.y ; 
		H.z = t * normal.z + ts * cosphi * plane_vec1.z + ts * sinphi * plane_vec2.z ; 
		Vector3D glossy_ray_direction ;
		glossy_ray_direction.x = 2 * t * H.x - view.x ;
		glossy_ray_direction.y = 2 * t * H.y - view.y ;
		glossy_ray_direction.z = 2 * t * H.z - view.z ;

		Vector3D point_above_surface( intersection.x + DELTA * glossy_ray_direction.x, intersection.y + DELTA * glossy_ray_direction.y, intersection.z + DELTA * glossy_ray_direction.z ) ;
		u = view.dot( H ) ; Su = Fo + ( 1 - Fo ) * ( pow( 1 - u, 5 ) ) ;
		double vdash = fabs( glossy_ray_direction.dot( normal ) ) ;
		Ray glossy_ray( point_above_surface, glossy_ray_direction ) ;
		glossy_ray.n = original_n ;	
		// Compute directional factor
		double Gv = v / ( sigma * ( 1 - v ) + v ) ;
		double Gvdash = vdash / ( sigma * ( 1 - vdash ) + vdash ) ;
		double Zt = sigma / pow( ( 1 + sigma * t - t * t ), 2 ) ;
		double D = ( 1 - Gv * Gvdash * ( 1 - Zt ) ) / ( 4 * M_PI * vdash ) ;

		if( glossy_ray.trace() )
		{
			Color glossy_color = glossy_ray.shade( depth + 1 ) ;
			//printf("\nDebug: Adding glossy contribution %lf,%lf,%lf at depth %d from object %p\n", glossy_color.red, glossy_color.green, glossy_color.blue, depth + 1, glossy_ray.object ) ;
			total_red += Su * vdash * D * glossy_color.red ;
			total_green += Su * vdash * D * glossy_color.green ; 
			total_blue += Su * vdash * D * glossy_color.blue ;
			if( total_red > 1.0 ) total_red = 1.0 ;
			if( total_green > 1.0 ) total_green = 1.0 ;
			if( total_blue > 1.0 ) total_blue = 1.0 ;
			return Color( total_red, total_green, total_blue ) ;
		}
		else
		{
			total_red += Su * vdash * D * background.red ;
			total_green += Su * vdash * D * background.green ; 
			total_blue += Su * vdash * D * background.blue ;
			if( total_red > 1.0 ) total_red = 1.0 ;
			if( total_green > 1.0 ) total_green = 1.0 ;
			if( total_blue > 1.0 ) total_blue = 1.0 ;
			return Color( total_red, total_green, total_blue ) ;
		}
	}	
	return Color( 0, 0, 0 ) ;
}

Color SchlickSurface::texture_shade( double u, double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n, bool inside, int depth )
{
	// Nice try
	return shade( intersection, normal, view, original_n, inside, depth ) ;
}
