// class definition for Surface class - represents the generic optical properties of a surface

#ifndef __SURFACE_H
#define __SURFACE_H

#include "color.h"
#include "vector3D.h"
#include "ray.h"
#include "object.h"
#include "light.h"
#include "texture.h"

#include <math.h>
#include <vector>


#define DELTA 0.001f
#define I255 1.0/255.0

using std::vector ;

/*
 * Generic surface
 */
class Surface
{
	public:
		
		Texture* texture ; // A texture mapped onto this surface if any
		Surface() ;
		Surface( const Surface& ) ;	
		virtual Color shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
		virtual Color texture_shade( const double u, const double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
		Vector3D refract( double cos_r, double sin_i, double cos_i, const Vector3D& normal, const Vector3D& view );
		void setTexture( Texture* t ) ;
};

class PhongSurface : public Surface
{
	public:
	Color intrinsic_color ; // The intrinsic color (r,g,b) of a surface
	// constants for the Phong illumination model
	double ka , // ambient coefficient
	kd , // diffuse coefficient
	ks , // specular coefficient
	ns ; // Phong shininess - The exponent of cos a in the Phong specular reflection equation

	double kt , // transmission coefficient ( refractance )
	kr , // reflection coefficient ( reflectance )
	n  ; // refractive index of the object

	PhongSurface( Color, double, double, double, double, double, double, double ) ;
	PhongSurface( const PhongSurface& ) ;
	virtual Color shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
	virtual Color texture_shade( const double u, const double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;

};

// Simple light source
class EmissiveSurface : public Surface
{
	public:
	// Very simple
	double red, green, blue ; // Light color
	
	EmissiveSurface( Color ) ;
	EmissiveSurface( const EmissiveSurface& ) ;
	virtual Color shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
	virtual Color texture_shade( const double u, const double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
};	

// More complex - surface utilizing Schlick's reflectance model for dielectrics
// Works only if path tracing is enabled 
// This is because the implementing methods use Russian roulette
class SchlickSurface : public Surface
{
	public:
	
	double Fo ; // The specular reflectance at normal incidence
	double sigma ; // The roughness parameter for this surface
	
	double kr, kt, n ; // Reflection/Refraction parameters for specular surfaces
	
	// Derived parameters for russian roulette
	double g, d, s ;

	SchlickSurface( double, double, double, double, double ) ;
	SchlickSurface( const SchlickSurface& ) ;
	virtual Color shade( const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
	virtual Color texture_shade( const double u, const double v, const Vector3D& intersection, const Vector3D& normal, const Vector3D& view, const double original_n = 1, bool inside = false, int depth = 0 ) ;
};

double rand_double( void ) ;

#endif
