// triangle.cpp - Implements intersection and shading methods for a triangle
// Uses: the MT algorithm for ray-triangle intersection
// Reference: http://www.acm.org/jgt/papers/MollerTrumbore97/


#include "triangle.h"
#include "surface.h" // For definition of DELTA
#include "csg.h"
#include "vector3D.h"
#include "bvolume.h"

#include <math.h>
#include <map>

using std::map ;
extern map<Object*,int> objectMap ;

// Constructor - assume v1, v2, v3 are in correct (anticlockwise) order
// This form of triangle assumes flat shading
Triangle::Triangle( Surface* s, const Vector3D& v1, const Vector3D& v2, const Vector3D& v3 ) : Primitive(s)
{
	vertex[0] = v1 ; vertex[1] = v2 ; vertex[2] = v3 ;
	shading_model = SHADE_FLAT ;

	// Some precomputation for the MT algorithm
	edge1 = Vector3D( v2.x - v1.x, v2.y - v1.y, v2.z - v1.z ) ;
	edge2 = Vector3D( v3.x - v1.x, v3.y - v1.y, v3.z - v1.z ) ;
	normal = Vector3D( -edge1.x, -edge1.y, -edge1.z ) ;
	normal = normal.cross( edge2 ) ;
}

// More sophisticated phong shading for triangles - implemented by barycentric interpolation
Triangle::Triangle( Surface* s, const Vector3D& v1, const Vector3D& v2, const Vector3D& v3, const Vector3D& n1, const Vector3D& n2, const Vector3D& n3 ) : Primitive(s)
{
	vertex[0] = v1 ; vertex[1] = v2 ; vertex[2] = v3 ;
	n[0] = n1 ; n[1] = n2 ; n[2] = n3 ;
	shading_model = SHADE_PHONG ;
	
	// Some precomputation for the MT algorithm
	edge1 = Vector3D( v2.x - v1.x, v2.y - v1.y, v2.z - v1.z ) ;
	edge2 = Vector3D( v3.x - v1.x, v3.y - v1.y, v3.z - v1.z ) ;
	normal = Vector3D( -edge1.x, -edge1.y, -edge1.z ) ;
	normal = normal.cross( edge2 ) ;
}

// Copy constructor
Triangle::Triangle( const Triangle& t ) : Primitive()
{
	int i, j ;
	surface = t.surface ; shading_model = t.shading_model ;
	for( i = 0 ; i < 3 ; i++ )
		vertex[i] = t.vertex[i] ;
	normal = t.normal ;

	// Some precomputation for the MT algorithm
	edge1 = t.edge1 ; edge2 = t.edge2 ; normal = t.normal ;

	if( t.shading_model == SHADE_PHONG ) {
		for( i = 0 ; i < 3 ; i++ )
			n[i] = t.n[i] ;
	}

	isTransformed = t.isTransformed ;
	if( t.isTransformed ) {
		// Copy elements of transformation matrix one by one
		M = new double* [4] ;
		for( i = 0 ; i < 4 ; i++ )
			M[i] = new double[4] ;
		N = new double* [3] ;
		for( i = 0 ; i < 3 ; i++ )
			N[i] = new double[3] ;
		for( i = 0 ; i < 4 ; i++ )
			for( j = 0 ; j < 4 ; j++ )
			{
				M[i][j] = t.M[i][j] ;
				if( i != 3 && j != 3 ) N[i][j] = t.N[i][j] ;
			}
	}
}

// Quicker method for shadow rays
bool Triangle::shadow_intersect( Ray& r )
{
	Ray ray ;
	double det, det_reciprocal, u, v, new_length ;
	
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;
	
	Vector3D displacement( ray.origin.x - vertex[0].x, ray.origin.y - vertex[0].y, ray.origin.z - vertex[0].z ) ;
	det = ray.direction.dot( normal ) ;

	if( isZero( det ) )
		return false ;

	det_reciprocal = 1 / det ;
	
	u = det_reciprocal * ray.direction.dot( ( Vector3D(edge2) ).cross( displacement ) );
	if( u < 0.0 || u > 1.0 ) return false ;
	v = -det_reciprocal * ray.direction.dot( ( Vector3D(edge1) ).cross( displacement ) );
	if( v < 0.0 || ( u + v > 1.0 ) ) return false ;
	
	new_length = -det_reciprocal * displacement.dot( normal ) ;
	if( ray.length < new_length || new_length < 0 ) return false ;
	return true ;
}

// The crucial intersect() method - implements ray object intersection method for a triangle
// Determines if this triangle is the first along the line of sight from the eye to intersect the given ray
// If not, nothing is done otherwise the ray's object is updated to this triangle, which could of course again be
// overridden by another, closer object.
// We need to do all we can to speed up computation
// Uses the MT algorithm (Moller-Trumbore '97)
bool Triangle::intersect( Ray& r )
{
	Ray ray ;
	double det, det_reciprocal, u, v, new_length ;
	
	// Code added later to transform objects
	if( isTransformed ) {
		ray = transform_ray( r ) ;
	}

	else ray = r ;
	
	Vector3D displacement( ray.origin.x - vertex[0].x, ray.origin.y - vertex[0].y, ray.origin.z - vertex[0].z ) ;
	det = ray.direction.dot( normal ) ;

	if( isZero( det ) )
		return false ;

	det_reciprocal = 1 / det ;
	
	u = det_reciprocal * ray.direction.dot( ( Vector3D(edge2) ).cross( displacement ) );
	if( u < 0.0 || u > 1.0 ) return false ;
	v = -det_reciprocal * ray.direction.dot( ( Vector3D(edge1) ).cross( displacement ) );
	if( v < 0.0 || ( u + v > 1.0 ) ) return false ;

	new_length = -det_reciprocal * displacement.dot( normal ) ; 

	if( ray.length > new_length && new_length > 0 ) {
		ray.length = new_length ;
		r.object = ray.object = this ;
		r.inside = ray.inside = false ;
		r.u = u ; r.v = v ;		
		if( isTransformed ) {
		Vector3D point_of_intersection( ray.origin.x + ray.direction.x * ray.length,
						ray.origin.y + ray.direction.y * ray.length,
						ray.origin.z + ray.direction.z * ray.length 
					);
		Vector3D actual_point = transform_point( point_of_intersection ) ;
		r.length = ( Vector3D( actual_point.x - r.origin.x, 
				     actual_point.y - r.origin.y,
				     actual_point.z - r.origin.z 
				     ).length()
			   ) ;
		}

		else r.length = ray.length ;
	
		//printf("\nDebug: Ray intersects triangle %d with length = %lf origin = (%lf,%lf,%lf) normal = (%lf,%lf,%lf)\n", objectMap[ this ], r.length, r.origin.x, r.origin.y, r.origin.z, normal.x, normal.y, normal.z ) ;
		return true ;
	}

	return false ;
}

// Object shader for a triangle
// A little more complex than other object shaders in this ray tracer
// This is principally due to the additional code for Phong interpolation of the triangle's vertex normals
Color Triangle::shade( const Ray& ray, int depth )
{
	// Assume ray has already been intersected with the plane - otherwise this will NOT be called
	Vector3D intersection(	ray.origin.x + ( ray.length * ray.direction.x ),
			      	ray.origin.y + ( ray.length * ray.direction.y ),
				ray.origin.z + ( ray.length * ray.direction.z )
			);
	
	Vector3D view( -ray.direction.x, -ray.direction.y, -ray.direction.z ) ;
	
	Vector3D surface_normal ;

	if( shading_model == SHADE_FLAT ) 
		surface_normal = this -> normal ;

	else /* if( shading_model == SHADE_PHONG ) */ {
		// Interpolate the normal using u, v, 1 - u - v in triangle's coordinate system (BARYCENTRIC)
		double u = ray.u, v = ray.v ; 
		surface_normal.x = ( 1 - u - v ) * n[0].x + u * n[1].x + v * n[2].x ; 
		surface_normal.y = ( 1 - u - v ) * n[0].y + u * n[1].y + v * n[2].y ; 
		surface_normal.z = ( 1 - u - v ) * n[0].z + u * n[1].z + v * n[2].z ; 
	}
	
	if( isTransformed ) 
		surface_normal = transform_normal( surface_normal ) ;
	
	// Call surface shader with appropriate parameters
	return surface -> shade( intersection, Vector3D::normalize(surface_normal), view, ray.n, ray.inside, depth ) ;
}

// It is incorrect to include triangles in a CSG object
// This routine is included to satisfy the compiler
vector <CSG_Interval*> Triangle::compute_intervals( const Ray& r ) 
{
	vector <CSG_Interval*> result ;
	return result ;
}


void Triangle::computeBounds( bVolume& b ) 
{
	b.xmin = min3( vertex[0].x, vertex[1].x, vertex[2].x ) ;
	b.ymin = min3( vertex[0].y, vertex[1].y, vertex[2].y ) ;
	b.zmin = min3( vertex[0].z, vertex[1].z, vertex[2].z ) ;
	b.xmax = max3( vertex[0].x, vertex[1].x, vertex[2].x ) ;
	b.ymax = max3( vertex[0].y, vertex[1].y, vertex[2].y ) ;
	b.zmax = max3( vertex[0].z, vertex[1].z, vertex[2].z ) ;
}
